"""
https://peps.python.org/pep-0249/

==

apilevel = '2.0'

paramstyle:
 qmark - Question mark style, e.g. ...WHERE name=?
 numeric - Numeric, positional style, e.g. ...WHERE name=:1
 named - Named style, e.g. ...WHERE name=:name
 format - ANSI C printf format codes, e.g. ...WHERE name=%s
 pyformat - Python extended format codes, e.g. ...WHERE name=%(name)s

Exception
| Warning
| Error
 | InterfaceError
 | DatabaseError
  | DataError
  | OperationalError
  | IntegrityError
  | InternalError
  | ProgrammingError
  | NotSupportedError

Date(year, month, day)
Time(hour, minute, second)
Timestamp(year, month, day, hour, minute, second)
DateFromTicks(ticks)
TimeFromTicks(ticks)
TimestampFromTicks(ticks)
Binary(string)

STRING type
BINARY type
NUMBER type
DATETIME type
ROWID type
"""
import dataclasses as dc
import enum

from .. import lang
from .params import ParamStyle


##


class DbapiDialect(enum.Enum):
    SQLITE = 'sqlite'
    MYSQL = 'mysql'
    POSTGRES = 'postgres'
    SNOWFLAKE = 'snowflake'


@dc.dataclass(frozen=True, kw_only=True)
class DbapiDriver:
    name: str
    dialect: DbapiDialect
    param_style: ParamStyle

    package_name: str | None = None
    module_name: str | None = None


class DbapiDrivers(lang.Namespace):

    ##
    # sqlite

    SQLITE3 = DbapiDriver(
        name='sqlite3',
        dialect=DbapiDialect.SQLITE,
        param_style=ParamStyle.QMARK,
    )

    SQLEAN = DbapiDriver(
        name='sqlean',
        dialect=DbapiDialect.SQLITE,
        param_style=ParamStyle.QMARK,
        package_name='sqlean.py',
    )

    DUCKDB = DbapiDriver(
        name='duckdb',
        dialect=DbapiDialect.SQLITE,
        param_style=ParamStyle.QMARK,
    )

    ##
    # mysql

    PYMYSQL = DbapiDriver(
        name='pymysql',
        dialect=DbapiDialect.MYSQL,
        param_style=ParamStyle.PYFORMAT,
    )

    MYSQL = DbapiDriver(
        name='mysql',
        dialect=DbapiDialect.MYSQL,
        param_style=ParamStyle.PYFORMAT,
        package_name='mysql-connector-python',
        module_name='mysql.connector',
    )

    MYSQLCLIENT = DbapiDriver(
        name='mysqlclient',
        dialect=DbapiDialect.MYSQL,
        param_style=ParamStyle.FORMAT,
        package_name='mysqlclient',
        module_name='MySQLdb',
    )

    ##
    # postgres

    PG8000 = DbapiDriver(
        name='pg8000',
        dialect=DbapiDialect.POSTGRES,
        param_style=ParamStyle.FORMAT,
    )

    PSYCOPG2 = DbapiDriver(
        name='psycopg2',
        dialect=DbapiDialect.POSTGRES,
        param_style=ParamStyle.PYFORMAT,
    )

    PSYCOPG = DbapiDriver(
        name='psycopg',
        dialect=DbapiDialect.POSTGRES,
        param_style=ParamStyle.PYFORMAT,
    )

    ##
    # snowflake

    SNOWFLAKE = DbapiDriver(
        name='snowflake',
        dialect=DbapiDialect.SNOWFLAKE,
        param_style=ParamStyle.PYFORMAT,
        package_name='snowflake-connector-python',
        module_name='snowflake.connector',
    )



# ruff: noqa: UP007 UP045
import dataclasses as dc
import typing as ta

from .. import check
from .. import lang
from .values import ScalarTypedValue
from .values import TypedValue
from .values import UniqueTypedValue


if ta.TYPE_CHECKING:
    from . import collection as tvc
else:
    tvc = lang.proxy_import('.collection', __package__)


TypedValueT = ta.TypeVar('TypedValueT', bound=TypedValue)
TypedValueU = ta.TypeVar('TypedValueU', bound=TypedValue)

UniqueTypedValueU = ta.TypeVar('UniqueTypedValueU', bound=UniqueTypedValue)


##


@dc.dataclass()
class UnconsumedTypedValuesError(Exception):
    values: ta.Sequence[TypedValue]


class _NOT_SET(lang.Marker):  # noqa
    pass


class TypedValuesConsumer(ta.Generic[TypedValueT]):
    def __init__(
            self,
            src: ta.Union[
                'tvc.TypedValues',
                ta.Iterable[tuple[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]],
                ta.Mapping[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]],
            ],
    ) -> None:
        super().__init__()

        kvs: ta.Iterable[tuple[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]]
        if isinstance(src, tvc.TypedValues):
            kvs = src.items()
        elif isinstance(src, ta.Mapping):
            kvs = src.items()
        else:
            kvs = src

        dct: dict = {}
        for k, v in kvs:
            check.not_in(k, dct)
            dct[k] = v

        self._dct: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]] = dct

    #

    def check_empty(self) -> None:
        if bool(self._dct):
            raise UnconsumedTypedValuesError(list(self))

    def __enter__(self) -> 'TypedValuesConsumer[TypedValueT]':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.check_empty()

    #

    def __iter__(self) -> ta.Iterator[TypedValueT]:
        for v in self._dct.values():
            if isinstance(v, tuple):
                yield from v
            else:
                yield v

    def __len__(self) -> int:
        return len(self._dct)

    def __bool__(self) -> bool:
        return bool(self._dct)

    #

    def keys(self) -> ta.KeysView[type[TypedValueT]]:
        return self._dct.keys()

    def values(self) -> ta.ValuesView[TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.values()

    def items(self) -> ta.ItemsView[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.items()

    #

    @ta.overload
    def pop(
            self,
            tv: UniqueTypedValueU,
    ) -> UniqueTypedValueU:
        ...

    @ta.overload
    def pop(
            self,
            cls: type[UniqueTypedValueU],
            /,
            default: UniqueTypedValueU,
    ) -> UniqueTypedValueU:
        ...

    @ta.overload
    def pop(
            self,
            cls: type[UniqueTypedValueU],
            /,
            default: UniqueTypedValueU | None,
    ) -> UniqueTypedValueU | None:
        ...

    @ta.overload
    def pop(
            self,
            cls: type[TypedValueU],
    ) -> ta.Sequence[TypedValueU]:
        ...

    @ta.overload
    def pop(
            self,
            cls: type[TypedValueU],
            /,
            default: ta.Iterable[TypedValueU],
    ) -> ta.Sequence[TypedValueU]:
        ...

    def pop(self, key, /, default=_NOT_SET):
        if not isinstance(key, type):
            if default is not _NOT_SET:
                raise RuntimeError('Must not provide both an instance key and a default')
            default = key
            key = type(default)

        if (
                issubclass(key, UniqueTypedValue) and
                (utvc := key._unique_typed_value_cls) is not key  # noqa
        ):
            key = utvc

        try:
            return self._dct.pop(key)
        except KeyError:
            if default is not _NOT_SET:
                return default
            else:
                raise

    #

    def pop_scalar_kwargs(self, **kwargs: type[TypedValueT]) -> dict[str, TypedValueT]:
        dct: dict[str, TypedValueT] = {}
        for k, vc in kwargs.items():
            check.issubclass(vc, ScalarTypedValue)
            try:
                v = self.pop(vc)
            except KeyError:
                continue
            dct[k] = v.v  # type: ignore[attr-defined]
        return dct


def consume(
    *tvs: TypedValueT,
    override: bool = False,
    check_type: type | tuple[type, ...] | None = None,
) -> TypedValuesConsumer[TypedValueT]:
    return tvc.TypedValues(
        *tvs,
        override=override,
        check_type=check_type,
    ).consume()



import abc
import dataclasses as dc
import typing as ta

from ... import lang


##


@dc.dataclass(frozen=True)
class Node(lang.Abstract):
    @property
    @abc.abstractmethod
    def children(self) -> ta.Sequence['Node']:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class LeafNode(Node, lang.Abstract):
    @property
    def children(self) -> ta.Sequence[Node]:
        return []


##


UnaryArithmeticOperator: ta.TypeAlias = ta.Literal[
    'plus',
    'minus',
]


@dc.dataclass(frozen=True)
class ArithmeticUnary(Node, lang.Final):
    operator: UnaryArithmeticOperator
    expression: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.expression]


ArithmeticOperator: ta.TypeAlias = ta.Literal[
    'div',
    'divide',
    'minus',
    'modulo',
    'multiply',
    'plus',
]


@dc.dataclass(frozen=True)
class Arithmetic(Node, lang.Final):
    operator: ArithmeticOperator
    left: Node
    right: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right]


@dc.dataclass(frozen=True)
class Assign(Node, lang.Final):
    name: str
    expr: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.expr]


ComparatorName: ta.TypeAlias = ta.Literal[
    'eq',
    'ne',
    'lt',
    'gt',
    'lte',
    'gte',
]


@dc.dataclass(frozen=True)
class Comparator(Node, lang.Final):
    name: ComparatorName
    first: Node
    second: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.first, self.second]


@dc.dataclass(frozen=True)
class CurrentNode(LeafNode, lang.Final):
    pass


@dc.dataclass(frozen=True)
class RootNode(LeafNode, lang.Final):
    pass


@dc.dataclass(frozen=True)
class Expref(Node, lang.Final):
    expression: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.expression]


@dc.dataclass(frozen=True)
class FunctionExpression(Node, lang.Final):
    name: str
    args: ta.Sequence[Node]

    @property
    def children(self) -> ta.Sequence[Node]:
        return tuple(self.args)


@dc.dataclass(frozen=True)
class Field(LeafNode, lang.Final):
    name: str


@dc.dataclass(frozen=True)
class FilterProjection(Node, lang.Final):
    left: Node
    right: Node
    comparator: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right, self.comparator]


@dc.dataclass(frozen=True)
class Flatten(Node, lang.Final):
    node: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.node]


@dc.dataclass(frozen=True)
class Identity(LeafNode, lang.Final):
    pass


@dc.dataclass(frozen=True)
class Index(LeafNode, lang.Final):
    index: int


@dc.dataclass(frozen=True)
class IndexExpression(Node, lang.Final):
    nodes: ta.Sequence[Node]

    @property
    def children(self) -> ta.Sequence[Node]:
        return tuple(self.nodes)


@dc.dataclass(frozen=True)
class KeyValPair(Node, lang.Final):
    key_name: str
    node: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.node]


@dc.dataclass(frozen=True)
class LetExpression(Node, lang.Final):
    bindings: ta.Sequence[Node]
    expr: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [*self.bindings, self.expr]


@dc.dataclass(frozen=True)
class Literal(LeafNode, lang.Final):
    literal_value: ta.Any


@dc.dataclass(frozen=True)
class MultiSelectDict(Node, lang.Final):
    nodes: ta.Sequence[KeyValPair]

    @property
    def children(self) -> ta.Sequence[Node]:
        return tuple(self.nodes)


@dc.dataclass(frozen=True)
class MultiSelectList(Node, lang.Final):
    nodes: ta.Sequence[Node]

    @property
    def children(self) -> ta.Sequence[Node]:
        return tuple(self.nodes)


@dc.dataclass(frozen=True)
class OrExpression(Node, lang.Final):
    left: Node
    right: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right]


@dc.dataclass(frozen=True)
class AndExpression(Node, lang.Final):
    left: Node
    right: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right]


@dc.dataclass(frozen=True)
class NotExpression(Node, lang.Final):
    expr: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.expr]


@dc.dataclass(frozen=True)
class Pipe(Node, lang.Final):
    left: Node
    right: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right]


@dc.dataclass(frozen=True)
class Projection(Node, lang.Final):
    left: Node
    right: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right]


@dc.dataclass(frozen=True)
class Subexpression(Node, lang.Final):
    children_nodes: ta.Sequence[Node]

    @property
    def children(self) -> ta.Sequence[Node]:
        return tuple(self.children_nodes)


@dc.dataclass(frozen=True)
class Slice(LeafNode, lang.Final):
    start: int | None
    end: int | None
    step: int | None


@dc.dataclass(frozen=True)
class TernaryOperator(Node, lang.Final):
    condition: Node
    if_truthy: Node
    if_falsy: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.condition, self.if_truthy, self.if_falsy]


@dc.dataclass(frozen=True)
class ValueProjection(Node, lang.Final):
    left: Node
    right: Node

    @property
    def children(self) -> ta.Sequence[Node]:
        return [self.left, self.right]


@dc.dataclass(frozen=True)
class VariableRef(LeafNode, lang.Final):
    name: str



import abc
import dataclasses as dc
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import struct as _struct
else:
    _struct = lang.proxy_import('struct')


##


F = ta.TypeVar('F')
T = ta.TypeVar('T')
U = ta.TypeVar('U')


class FnPair(abc.ABC, ta.Generic[F, T]):
    @abc.abstractmethod
    def forward(self, f: F) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def backward(self, t: T) -> F:
        raise NotImplementedError

    ##

    def __call__(self, f: F) -> T:
        return self.forward(f)

    def invert(self) -> 'FnPair[T, F]':
        if isinstance(self, Inverted):
            return self.fp
        return Inverted(self)

    def compose(self, nxt: 'FnPair[T, U]') -> 'FnPair[F, U]':
        return Composite((self, nxt))


##


@lang.unabstract_class(['forward', 'backward'])
@dc.dataclass(frozen=True)
class Simple(FnPair[F, T]):
    forward: ta.Callable[[F], T]  # type: ignore
    backward: ta.Callable[[T], F]  # type: ignore


of = Simple

NOP: FnPair[ta.Any, ta.Any] = of(lang.identity, lang.identity)


##


@dc.dataclass(frozen=True)
class Inverted(FnPair[F, T]):
    fp: FnPair[T, F]

    def forward(self, f: F) -> T:
        return self.fp.backward(f)

    def backward(self, t: T) -> F:
        return self.fp.forward(t)


##


@dc.dataclass(frozen=True)
class Composite(FnPair[F, T]):
    children: ta.Sequence[FnPair]

    def forward(self, f: F) -> T:
        for c in self.children:
            f = c.forward(f)
        return ta.cast(T, f)

    def backward(self, t: T) -> F:
        for c in reversed(self.children):
            t = c.backward(t)
        return ta.cast(F, t)


I0 = ta.TypeVar('I0')
I1 = ta.TypeVar('I1')
I2 = ta.TypeVar('I2')
I3 = ta.TypeVar('I3')
I4 = ta.TypeVar('I4')


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, I2],
        fp3: FnPair[I2, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, I2],
        fp3: FnPair[I2, I3],
        fp4: FnPair[I3, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(
        fp0: FnPair[F, I0],
        f01: FnPair[I0, I1],
        fp2: FnPair[I1, I2],
        fp3: FnPair[I2, I3],
        fp4: FnPair[I3, I4],
        fp5: FnPair[I4, T],
) -> FnPair[F, T]:
    ...


@ta.overload
def compose(*ps: FnPair) -> FnPair:
    ...


def compose(*ps):
    if not ps:
        return NOP
    if len(ps) == 1:
        return ps[0]
    return Composite(ps)


##


@dc.dataclass(frozen=True)
class Optional(FnPair[F | None, T | None]):
    fp: FnPair[F, T]

    def forward(self, f: F | None) -> T | None:
        return None if f is None else self.fp.forward(f)

    def backward(self, t: T | None) -> F | None:
        return None if t is None else self.fp.backward(t)


class Lines(FnPair[ta.Sequence[str], str]):
    def forward(self, f: ta.Sequence[str]) -> str:
        return '\n'.join(f)

    def backward(self, t: str) -> ta.Sequence[str]:
        return t.splitlines()


##


@dc.dataclass(frozen=True)
class Struct(FnPair[tuple, bytes]):
    fmt: str

    def forward(self, f: tuple) -> bytes:
        return _struct.pack(self.fmt, *f)

    def backward(self, t: bytes) -> tuple:
        return _struct.unpack(self.fmt, t)


##


Object: ta.TypeAlias = FnPair[ta.Any, T]
ObjectStr: ta.TypeAlias = Object[str]
ObjectBytes: ta.TypeAlias = Object[bytes]


class Object_(FnPair[ta.Any, T], lang.Abstract):  # noqa
    pass


class ObjectStr_(Object_[str], lang.Abstract):  # noqa
    pass


class ObjectBytes_(Object_[bytes], lang.Abstract):  # noqa
    pass



# ruff: noqa: PYI055 UP007
"""
TODO:
 - \u0123 in strings
 - https://clojure.org/reference/reader
 - reader meta - ^:foo
 - read table
"""
import dataclasses as dc
import datetime
import enum
import io
import re
import typing as ta

from ... import check
from ...funcs.genmachine import GenMachine
from .lexing import Position
from .lexing import StreamLexer
from .lexing import Token
from .values import Char
from .values import Collection
from .values import Keyword
from .values import List
from .values import Map
from .values import Set
from .values import Symbol
from .values import Tagged
from .values import Vector


##


WORD_CONST_VALUES: ta.Mapping[str, ta.Any] = {
    'true': True,
    'false': False,
    'nil': None,

    '##Inf': float('inf'),
    '##-Inf': float('-inf'),
    '##NaN': float('nan'),
}


STRING_ESCAPE_MAP: ta.Mapping[str, str] = {
    '"': '"',
    '\\': '\\',
    'b': '\b',
    'f': '\f',
    'n': '\n',
    'r': '\r',
    't': '\t',
}


CHAR_ESCAPE_MAP: ta.Mapping[str, str] = {
    'backspace': '\b',
    'formfeed': '\f',
    'newline': '\n',
    'return': '\r',
    'space': ' ',
    'tab': '\t',
}


##


@dc.dataclass()
class StreamParseError(Exception):
    message: str

    pos: Position | None = None


@dc.dataclass(frozen=True)
class MetaMaker:
    fn: ta.Callable[..., ta.Any]

    def __call__(self, *args: ta.Any, meta: ta.Any | None = None) -> ta.Any:
        return self.fn(*args, meta=meta)


class StreamParser(GenMachine[Token, ta.Any]):
    DEFAULT_TAG_HANDLERS: ta.ClassVar[ta.Mapping[str, ta.Callable[..., ta.Any]]] = {
        'inst': lambda val: datetime.datetime.fromisoformat(val) if isinstance(val, str) else None,
    }

    def __init__(
            self,
            *,
            keyword_maker: ta.Callable[..., ta.Any] = MetaMaker(Keyword),
            char_maker: ta.Callable[..., ta.Any] = MetaMaker(Char),
            symbol_maker: ta.Callable[..., ta.Any] = MetaMaker(Symbol),

            list_maker: ta.Callable[..., ta.Any] = MetaMaker(List.new),
            vector_maker: ta.Callable[..., ta.Any] = MetaMaker(Vector.new),
            set_maker: ta.Callable[..., ta.Any] = MetaMaker(Set.new),
            map_maker: ta.Callable[..., ta.Any] = MetaMaker(Map.new),

            tag_handlers: ta.Mapping[str, ta.Callable[..., ta.Any]] | None = None,
    ) -> None:
        self._keyword_maker = keyword_maker
        self._char_maker = char_maker
        self._symbol_maker = symbol_maker

        self._list_maker = list_maker
        self._vector_maker = vector_maker
        self._set_maker = set_maker
        self._map_maker = map_maker

        self._tag_handlers = {
            **self.DEFAULT_TAG_HANDLERS,
            **(tag_handlers or {}),
        }

        self._stack: list[
            tuple[
                ta.Union[
                    type[Collection],
                    StreamParser._StackSpecial,
                ],
                list[ta.Any],
            ],
        ] = []

        super().__init__(self._do_main())

    class _StackSpecial(enum.Enum):  # noqa
        DISCARD = enum.auto()
        TAG = enum.auto()

    def _emit_value(self, value: ta.Any) -> tuple[ta.Any, ...]:
        while self._stack and self._stack[-1][0] is StreamParser._StackSpecial.TAG:
            cc, cl = self._stack.pop()
            ts = check.non_empty_str(check.single(cl))
            value = Tagged(ts, value)

        if not self._stack:
            return (value,)

        cc, cl = self._stack[-1]

        if cc is StreamParser._StackSpecial.DISCARD:
            check.empty(cl)
            self._stack.pop()
            return ()

        elif cc is StreamParser._StackSpecial.TAG:
            ts = check.non_empty_str(check.single(cl))
            self._stack.pop()
            tv = Tagged(ts, value)
            return (tv,)

        elif cc is Map:
            if cl and len(cl[-1]) < 2:
                cl[-1] = (*cl[-1], value)
            else:
                cl.append((value,))
            return ()

        elif isinstance(cc, type) and issubclass(cc, Collection):
            cl.append(value)
            return ()

        else:
            raise RuntimeError(cc)

    def _do_main(self):
        while True:
            tok: Token
            try:
                tok = yield None  # noqa
            except GeneratorExit:
                if self._stack:
                    raise StreamParseError('Expected value') from None
                else:
                    raise

            # ignored

            if tok.kind in ('SPACE', 'COMMENT'):
                continue

            # scalars

            value: ta.Any

            if tok.kind == 'STRING':
                value = self._parse_string(tok)

            elif tok.kind == 'CHAR':
                value = self._parse_char(tok)

            elif tok.kind == 'WORD':
                if tok.src.startswith('#'):
                    # FIXME: more dispatching
                    self._stack.append((StreamParser._StackSpecial.TAG, [tok.src[1:]]))
                    continue

                else:
                    value = self._parse_word(tok)

            # open

            elif tok.kind == 'LPAREN':
                self._stack.append((List, []))
                continue

            elif tok.kind == 'LBRACKET':
                self._stack.append((Vector, []))
                continue

            elif tok.kind == 'HASH_LBRACE':
                self._stack.append((Set, []))
                continue

            elif tok.kind == 'LBRACE':
                self._stack.append((Map, []))
                continue

            elif tok.kind == 'HASH_UNDERSCORE':
                self._stack.append((StreamParser._StackSpecial.DISCARD, []))
                continue

            # close

            elif tok.kind == 'RPAREN':
                cc, cl = self._stack.pop()
                check.state(cc is List)
                value = self._list_maker(cl)

            elif tok.kind == 'RBRACKET':
                cc, cl = self._stack.pop()
                check.state(cc is Vector)
                value = self._vector_maker(cl)

            elif tok.kind == 'RBRACE':
                cc, cl = self._stack.pop()

                if cc is Set:
                    value = self._set_maker(cl)

                elif cc is Map:
                    if cl and len(cl[-1]) != 2:
                        raise RuntimeError('Mismatched map entries')
                    value = self._map_maker(cl)

                else:
                    raise RuntimeError(cc)

            # nyi

            elif tok.kind == 'META':
                raise NotImplementedError

            elif tok.kind == 'QUOTE':
                raise NotImplementedError

            # failure

            else:
                raise ValueError(tok.kind)

            # emit

            if (ev := self._emit_value(value)):
                yield ev

    def _parse_string(self, tok: Token) -> str:
        check.state(tok.kind == 'STRING')
        src = tok.src
        check.state(src[0] == '"')
        check.state(src[-1] == '"')
        check.state(len(src) > 1)

        p = 1
        end = len(src) - 1
        if src.find('\\', p, end) < 0:
            return src[1:-1]

        sb = io.StringIO()
        while True:
            if (n := src.find('\\', p, end)) < 0:
                sb.write(src[p:end])
                break

            sb.write(src[p:n])
            p = n + 1
            check.state(p < end)
            x = src[p]
            p += 1

            if x == 'u':
                check.state(p < end - 4)
                r = chr(int(src[p:p + 4], 16))
                p += 4

            else:
                r = STRING_ESCAPE_MAP[x]

            sb.write(r)

        return sb.getvalue()

    def _parse_char(self, tok: Token) -> ta.Any:
        check.state(tok.kind == 'CHAR')
        src = tok.src
        check.state(len(src) > 1)
        check.state(src.startswith('\\'))

        if len(src) == 2:
            c = src[1]

        elif src[1] == 'u':
            check.state(len(src) == 6)
            c = chr(int(src[2:], 16))  # noqa

        elif src[1] == 'o':
            # \oXXX -> octal
            raise NotImplementedError

        else:
            c = CHAR_ESCAPE_MAP[src[1:]]

        return self._char_maker(c)

    _INT_PAT = re.compile(r'[-+]?(0|[1-9][0-9]*)N?')
    _FLOAT_PAT = re.compile(r'[-+]?(0|[1-9][0-9]*)(\.[0-9]+)?([eE][+-]?(0|[1-9][0-9]*))?M?')

    def _parse_word(self, tok: Token) -> ta.Any:
        check.state(tok.kind == 'WORD')
        src = tok.src
        check.non_empty_str(src)
        check.state(not src.startswith('#'))

        if src in WORD_CONST_VALUES:
            return WORD_CONST_VALUES[src]

        elif src.startswith(':'):
            return self._keyword_maker(src[1:])

        elif self._INT_PAT.fullmatch(src):
            # FIXME: numbers lol
            # 2r101010, 052, 8r52, 0x2a, 36r16, and 42 are all the same Long.
            # Floating point numbers are read as Doubles; with M suffix they are read as BigDecimals.
            # Ratios are supported, e.g. 22/7.
            if src.endswith('N'):
                return int(src[:-1])

            else:
                return int(src)

        elif self._FLOAT_PAT.fullmatch(src):
            return float(src)

        else:
            return self._symbol_maker(src)


##


def parse_list(src: str, **kwargs: ta.Any) -> list[ta.Any]:
    r: list[ta.Any] = []
    with StreamLexer() as l:
        with StreamParser(**kwargs) as p:
            for c in [*src, '']:
                for t in l(c):
                    for o in p(t):
                        r.append(o)  # noqa
    return r


def parse(src: str, **kwargs: ta.Any) -> ta.Any | None:
    values = parse_list(src, **kwargs)
    if not values:
        return None
    return check.single(values)



import typing as ta


T = ta.TypeVar('T')


##


class _ClassOrInstanceMethod:
    def __init__(self, func):
        super().__init__()

        self.__func__ = func

    def __get__(self, instance, owner=None):
        return self.__func__.__get__(instance if instance is not None else owner)


class BindableClass(ta.Generic[T]):
    # FIXME: apparently can't have TypeVars in ClassVars, but could stick in a @classmethod (which gets transformed)...
    _bound: ta.ClassVar[type[T] | None] = None  # type: ignore[misc]

    def __init__(self, *, _bound):
        super().__init__()

        setattr(self, '_bound', _bound)

    def __class_getitem__(cls, *args, **kwargs):
        # FIXME: this could handle __mro_items__ to be subclassable, but it's not currently really intended to be
        #   subclassed
        if cls is BindableClass:
            return super().__class_getitem__(*args, **kwargs)  # type: ignore[misc]

        [bind_cls] = args
        return cls(_bound=bind_cls)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        for k, v in cls.__dict__.items():
            if isinstance(v, classmethod):
                setattr(cls, k, _ClassOrInstanceMethod(v.__func__))



# ruff: noqa: UP007 UP045
# @omlish-lite
"""
TODO:
 - 'json pids', with code version? '.json.pid'? '.jpid'?
  - json*L* pidfiles - first line is bare int, following may be json - now `head -n1 foo.pid` not cat
"""
import errno
import fcntl
import os
import signal
import typing as ta


##


class Pidfile:
    def __init__(
            self,
            path: str,
            *,
            inheritable: bool = True,
            no_create: bool = False,
    ) -> None:
        super().__init__()

        self._path = path
        self._inheritable = inheritable
        self._no_create = no_create

    @property
    def path(self) -> str:
        return self._path

    @property
    def inheritable(self) -> bool:
        return self._inheritable

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._path!r})'

    #

    _f: ta.TextIO

    def fileno(self) -> ta.Optional[int]:
        if hasattr(self, '_f'):
            return self._f.fileno()
        else:
            return None

    #

    _fd_to_dup: int

    def dup(self) -> 'Pidfile':
        fd = self._f.fileno()
        dup = Pidfile(
            self._path,
            inheritable=self._inheritable,
        )
        dup._fd_to_dup = fd  # noqa
        return dup

    #

    def __enter__(self) -> 'Pidfile':
        if hasattr(self, '_fd_to_dup'):
            fd = os.dup(self._fd_to_dup)
            del self._fd_to_dup

        else:
            ofl = os.O_RDWR
            if not self._no_create:
                ofl |= os.O_CREAT
            fd = os.open(self._path, ofl, 0o600)

        try:
            if self._inheritable:
                os.set_inheritable(fd, True)

            f = os.fdopen(fd, 'r+')

        except BaseException:
            os.close(fd)
            raise

        self._f = f
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    #

    def __getstate__(self):
        state = self.__dict__.copy()

        if '_f' in state:
            # self._inheritable may be decoupled from actual file inheritability - for example when using the manager.
            if os.get_inheritable(fd := state.pop('_f').fileno()):
                state['__fd'] = fd

        return state

    def __setstate__(self, state):
        if '_f' in state:
            raise RuntimeError

        if '__fd' in state:
            state['_f'] = os.fdopen(state.pop('__fd'), 'r+')

        self.__dict__.update(state)

    #

    def close(self) -> bool:
        if not hasattr(self, '_f'):
            return False

        self._f.close()
        del self._f
        return True

    def try_acquire_lock(self) -> bool:
        try:
            fcntl.flock(self._f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return True

        except BlockingIOError as e:
            if e.errno == errno.EAGAIN:
                return False
            else:
                raise

    #

    class Error(Exception):
        pass

    class LockedError(Error):
        pass

    def acquire_lock(self) -> None:
        if not self.try_acquire_lock():
            raise self.LockedError

    class NotLockedError(Error):
        pass

    def ensure_cannot_lock(self) -> None:
        if self.try_acquire_lock():
            raise self.NotLockedError

    #

    def write(
            self,
            pid: ta.Optional[int] = None,
            *,
            suffix: ta.Optional[str] = None,
    ) -> None:
        self.acquire_lock()

        if pid is None:
            pid = os.getpid()

        self._f.seek(0)
        self._f.truncate()
        self._f.write('\n'.join([
            str(pid),
            *([suffix] if suffix is not None else []),
            '',
        ]))
        self._f.flush()

    def clear(self) -> None:
        self.acquire_lock()

        self._f.seek(0)
        self._f.truncate()

    #

    def read_raw(self) -> ta.Optional[str]:
        self.ensure_cannot_lock()

        self._f.seek(0)
        buf = self._f.read()
        if not buf:
            return None
        return buf

    def read(self) -> ta.Optional[int]:
        buf = self.read_raw()
        if not buf:
            return None
        return int(buf.splitlines()[0].strip())

    def kill(self, sig: int = signal.SIGTERM) -> None:
        if (pid := self.read()) is None:
            raise self.Error(f'Pidfile locked but empty')
        os.kill(pid, sig)



"""
https://github.com/python/cpython/blob/21d2a9ab2f4dcbf1be462d3b7f7a231a46bc1cb7/Parser/asdl.py

-------------------------------------------------------------------------------

Parser for ASDL [1] definition files. Reads in an ASDL description and parses it into an AST that describes it.

The EBNF we're parsing here: Figure 1 of the paper [1]. Extended to support modules and attributes after a product.
Words starting with Capital letters are terminals. Literal tokens are in "double quotes". Others are non-terminals. Id
is either TokenId or ConstructorId.

module        ::= "module" Id "{" [definitions] "}"
definitions   ::= { TypeId "=" type }
type          ::= product | sum
product       ::= fields ["attributes" fields]
fields        ::= "(" { field, "," } field ")"
field         ::= TypeId ["?" | "*"] [Id]
sum           ::= constructor { "|" constructor } ["attributes" fields]
constructor   ::= ConstructorId [fields]

[1] "The Zephyr Abstract Syntax Description Language" by Wang, et. al. See http://asdl.sourceforge.net/
"""
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001-2024 Python Software Foundation; All Rights
# Reserved" are retained in Python alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
# ruff: noqa: N802
import abc
import dataclasses as dc
import enum
import re
import typing as ta

from .. import cached
from .. import check


##


# The following classes define nodes into which the Asdl description is parsed. Note: this is a "meta-AST". Asdl files
# (such as Python.asdl) describe the AST structure used by a programming language. But Asdl files themselves need to be
# parsed. This module parses Asdl files and uses a simple AST to represent them. See the EBNF at the top of the file to
# understand the logical connection between the various node types.

class Ast(abc.ABC):
    @abc.abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class Field(Ast):
    type: str
    name: str | None = None
    seq: bool = False
    opt: bool = False

    def __str__(self) -> str:
        if self.seq:
            extra = '*'
        elif self.opt:
            extra = '?'
        else:
            extra = ''

        return f'{self.type}{extra} {self.name}'

    def __repr__(self) -> str:
        if self.seq:
            extra = ', seq=True'
        elif self.opt:
            extra = ', opt=True'
        else:
            extra = ''
        if self.name is None:
            return f'Field({self.type}{extra})'
        else:
            return f'Field({self.type}, {self.name}{extra})'


@dc.dataclass(frozen=True)
class Constructor(Ast):
    name: str
    fields: ta.Sequence[Field] | None = None

    def __repr__(self) -> str:
        return f'Constructor({self.name}, {self.fields})'


@dc.dataclass(frozen=True)
class Sum(Ast):
    types: ta.Sequence[Constructor]
    attributes: ta.Sequence[Field] | None = None

    def __repr__(self) -> str:
        if self.attributes:
            return f'Sum({self.types}, {self.attributes})'
        else:
            return f'Sum({self.types})'


@dc.dataclass(frozen=True)
class Product(Ast):
    fields: ta.Sequence[Field]
    attributes: ta.Sequence[Field] | None = None

    def __repr__(self) -> str:
        if self.attributes:
            return f'Product({self.fields}, {self.attributes})'
        else:
            return f'Product({self.fields})'


@dc.dataclass(frozen=True)
class Type(Ast):
    name: str
    value: Sum | Product

    def __repr__(self) -> str:
        return f'Type({self.name}, {self.value})'


@dc.dataclass(frozen=True)
class Module(Ast):
    name: str
    dfns: ta.Sequence[Type]

    @cached.property
    def types(self) -> ta.Mapping[str, Sum | Product]:
        return {ty.name: ty.value for ty in self.dfns}

    def __repr__(self) -> str:
        return f'Module({self.name}, {self.dfns})'


##


# A generic visitor for the meta-Ast that describes Asdl. This can be used by emitters. Note that this visitor does not
# provide a generic visit method, so a subclass needs to define visit methods from visitModule to as deep as the
# interesting node.
# We also define a Check visitor that makes sure the parsed Asdl is well-formed.

class VisitorBase:
    """Generic tree visitor for Asts."""

    def __init__(self) -> None:
        super().__init__()
        self.cache: dict[type, ta.Any] = {}

    def visit(self, obj, *args):
        klass = obj.__class__
        meth = self.cache.get(klass)
        if meth is None:
            methname = 'visit' + klass.__name__
            meth = getattr(self, methname, None)
            self.cache[klass] = meth
        if meth:
            try:
                meth(obj, *args)
            except Exception as e:
                raise Exception(f'Error visiting {obj!r}') from e


##


BUILTIN_TYPES: ta.AbstractSet[str] = {
    'identifier',
    'string',
    'int',
    'constant',
}


##


# Types for describing tokens in an Asdl specification.
class TokenKind(enum.IntEnum):
    """TokenKind is provides a scope for enumerated token kinds."""

    CONSTRUCTOR_ID = enum.auto()
    TYPE_ID = enum.auto()
    EQUALS = enum.auto()
    COMMA = enum.auto()
    QUESTION = enum.auto()
    PIPE = enum.auto()
    ASTERISK = enum.auto()
    L_PAREN = enum.auto()
    R_PAREN = enum.auto()
    L_BRACE = enum.auto()
    R_BRACE = enum.auto()


OPERATOR_TABLE: ta.Mapping[str, TokenKind] = {
    '=': TokenKind.EQUALS,
    ',': TokenKind.COMMA,
    '?': TokenKind.QUESTION,
    '|': TokenKind.PIPE,
    '(': TokenKind.L_PAREN,
    ')': TokenKind.R_PAREN,
    '*': TokenKind.ASTERISK,
    '{': TokenKind.L_BRACE,
    '}': TokenKind.R_BRACE,
}


@dc.dataclass(frozen=True)
class Token:
    kind: TokenKind
    value: str
    lineno: int


class AsdlSyntaxError(Exception):
    def __init__(self, msg: str, lineno: int | None = None) -> None:
        super().__init__()
        self.msg = msg
        self.lineno = lineno or '<unknown>'

    def __str__(self) -> str:
        return f'Syntax error on line {self.lineno}: {self.msg}'


def tokenize_asdl(buf: str) -> ta.Iterator[Token]:
    """Tokenize the given buffer. Yield Token objects."""

    for lineno, line in enumerate(buf.splitlines(), 1):
        for m in re.finditer(r'\s*(\w+|--.*|.)', line.strip()):
            c = m.group(1)
            if c[0].isalpha():
                # Some kind of identifier
                if c[0].isupper():
                    yield Token(TokenKind.CONSTRUCTOR_ID, c, lineno)
                else:
                    yield Token(TokenKind.TYPE_ID, c, lineno)
            elif c[:2] == '--':
                # Comment
                break
            else:
                # Operators
                try:
                    op_kind = OPERATOR_TABLE[c]
                except KeyError:
                    raise AsdlSyntaxError(f'Invalid operator {c}', lineno)  # noqa
                yield Token(op_kind, c, lineno)


##


class AsdlParser:
    """
    Parser for Asdl files.

    Create, then call the parse method on a buffer containing Asdl. This is a simple recursive descent parser that uses
    tokenize_asdl for the lexing.
    """

    def __init__(self) -> None:
        super().__init__()
        self._tokenizer: ta.Iterator[Token] | None = None
        self.cur_token: Token | None = None

    def cur(self) -> Token:
        return check.not_none(self.cur_token)

    def parse(self, buf: str) -> Module:
        """Parse the Asdl in the buffer and return an Ast with a Module root."""

        self._tokenizer = tokenize_asdl(buf)
        self._advance()
        return self._parse_module()

    def _parse_module(self) -> Module:
        if self._at_keyword('module'):
            self._advance()
        else:
            raise AsdlSyntaxError(f'Expected "module" (found {self.cur().value})', self.cur().lineno)  # noqa
        name = self._match(self._id_kinds)
        self._match(TokenKind.L_BRACE)
        defs = self._parse_definitions()
        self._match(TokenKind.R_BRACE)
        return Module(name, defs)

    def _parse_definitions(self) -> ta.Sequence[Type]:
        defs = []
        while self.cur().kind == TokenKind.TYPE_ID:
            typename = check.non_empty_str(self._advance())
            self._match(TokenKind.EQUALS)
            ty = self._parse_type()
            defs.append(Type(typename, ty))
        return defs

    def _parse_type(self) -> Sum | Product:
        if self.cur().kind == TokenKind.L_PAREN:
            # If we see a (, it's a product
            return self._parse_product()
        else:
            # Otherwise it's a sum. Look for ConstructorId
            sumlist = [Constructor(self._match(TokenKind.CONSTRUCTOR_ID), self._parse_optional_fields())]
            while self.cur().kind == TokenKind.PIPE:
                # More constructors
                self._advance()
                sumlist.append(Constructor(
                    self._match(TokenKind.CONSTRUCTOR_ID),
                    self._parse_optional_fields()),
                )
            return Sum(sumlist, self._parse_optional_attributes())

    def _parse_product(self) -> Product:
        return Product(self._parse_fields(), self._parse_optional_attributes())

    def _parse_fields(self) -> ta.Sequence[Field]:
        fields = []
        self._match(TokenKind.L_PAREN)
        while self.cur().kind == TokenKind.TYPE_ID:
            typename = check.non_empty_str(self._advance())
            is_seq, is_opt = self._parse_optional_field_quantifier()
            id = self._advance() if self.cur().kind in self._id_kinds else None  # noqa
            fields.append(Field(typename, id, seq=is_seq, opt=is_opt))
            if self.cur().kind == TokenKind.R_PAREN:
                break
            elif self.cur().kind == TokenKind.COMMA:
                self._advance()
        self._match(TokenKind.R_PAREN)
        return fields

    def _parse_optional_fields(self) -> ta.Sequence[Field] | None:
        if self.cur().kind == TokenKind.L_PAREN:
            return self._parse_fields()
        else:
            return None

    def _parse_optional_attributes(self) -> ta.Sequence[Field] | None:
        if self._at_keyword('attributes'):
            self._advance()
            return self._parse_fields()
        else:
            return None

    def _parse_optional_field_quantifier(self) -> tuple[bool, bool]:  # (seq, opt)
        is_seq, is_opt = False, False
        if self.cur().kind == TokenKind.ASTERISK:
            is_seq = True
            self._advance()
        elif self.cur().kind == TokenKind.QUESTION:
            is_opt = True
            self._advance()
        return is_seq, is_opt

    def _advance(self) -> str | None:
        """Return the value of the current token and read the next one into self.cur_token."""

        cur_val = None if self.cur_token is None else self.cur_token.value
        try:
            self.cur_token = next(check.not_none(self._tokenizer))
        except StopIteration:
            self.cur_token = None
        return cur_val

    _id_kinds = (TokenKind.CONSTRUCTOR_ID, TokenKind.TYPE_ID)

    def _match(self, kind: TokenKind | tuple[TokenKind, ...]) -> str:
        """
        The 'match' primitive of RD parsers.

        * Verifies that the current token is of the given kind (kind can be a tuple, in which the kind must match one of
          its members).
        * Returns the value of the current token
        * Reads in the next token
        """

        if (isinstance(kind, tuple) and self.cur().kind in kind) or self.cur().kind == kind:
            value = self.cur().value
            self._advance()
            return value
        else:
            raise AsdlSyntaxError(
                f'Unmatched {kind} (found {self.cur().kind})',
                self.cur().lineno,
            )

    def _at_keyword(self, keyword: str) -> bool:
        return self.cur().kind == TokenKind.TYPE_ID and self.cur().value == keyword


##


FlatFieldArity: ta.TypeAlias = ta.Literal[1, '?', '*']


@dc.dataclass(frozen=True)
class FlatField:
    name: str
    type: str
    n: FlatFieldArity = 1


@dc.dataclass(frozen=True)
class FlatNode(abc.ABC):
    name: str
    fields: ta.Sequence[FlatField] = dc.field(default=(), kw_only=True)
    attributes: ta.Sequence[FlatField] = dc.field(default=(), kw_only=True)


@dc.dataclass(frozen=True)
class FlatSum(FlatNode):
    constructors: ta.Sequence[str] = dc.field(default=(), kw_only=True)


@dc.dataclass(frozen=True)
class FlatProduct(FlatNode):
    pass


@dc.dataclass(frozen=True)
class FlatConstructor(FlatNode):
    sum: str = dc.field(kw_only=True)


def flatten(mod: Module) -> ta.Mapping[str, FlatNode]:
    lst: list[FlatNode] = []

    def mk_field(af: Field) -> FlatField:
        return FlatField(
            check.non_empty_str(af.name),
            af.type,
            n='*' if af.seq else '?' if af.opt else 1,
        )

    def mk_fields(afs: ta.Iterable[Field] | None) -> ta.Sequence[FlatField]:
        return list(map(mk_field, afs or []))

    for ty in mod.dfns:
        v = ty.value

        if isinstance(v, Sum):
            lst.append(FlatSum(
                ty.name,
                attributes=mk_fields(v.attributes),
                constructors=[c.name for c in v.types],
            ))

            for c in v.types:
                lst.append(FlatConstructor(
                    c.name,
                    fields=mk_fields(c.fields),
                    attributes=mk_fields(v.attributes),
                    sum=ty.name,
                ))

        elif isinstance(v, Product):
            lst.append(FlatProduct(
                ty.name,
                fields=mk_fields(v.fields),
                attributes=mk_fields(v.attributes),
            ))

        else:
            raise TypeError(v)

    #

    dct: dict[str, FlatNode] = {}

    for n in lst:
        if n.name in dct:
            raise KeyError(n.name)
        dct[n.name] = n

    return dct


#


def build_fields_info(
        nodes: ta.Mapping[str, FlatNode],
        *,
        exclude_builtins: bool = False,
) -> ta.Mapping[str, ta.Mapping[str, tuple[str, FlatFieldArity]]]:
    dct: dict[str, dict[str, tuple[str, FlatFieldArity]]] = {}
    for n in nodes.values():
        cur = {}
        f: FlatField
        for f in [*n.fields, *n.attributes]:
            if f.type in BUILTIN_TYPES:
                if exclude_builtins:
                    continue
            elif f.type not in nodes:
                raise KeyError(f.type)
            cur[f.name] = (f.type, f.n)
        if cur:
            dct[n.name] = cur
    return dct



# Copyright (c) 2016-2021 Daniel Oaks
# Copyright (c) 2018-2021 Shivaram Lingamneni
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
# granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#
# https://github.com/ergochat/irc-go/blob/9beac2d29dc5f998c5a53e5db7a6426d7d083a79/ircmsg/
from .consts import (  # noqa
    MAX_LEN_CLIENT_TAG_DATA,
    MAX_LEN_SERVER_TAG_DATA,
    MAX_LEN_TAGS,
    MAX_LEN_TAGS_FROM_CLIENT,
    MAX_LEN_TAG_DATA,
)

from .errors import (  # noqa
    BadCharactersError,
    BadParamError,
    CommandMissingError,
    Error,
    InvalidTagContentError,
    LineEmptyError,
    MalformedNuhError,
    TagsTooLongError,
)

from .message import (  # noqa
    Message,
)

from .nuh import (  # noqa
    Nuh,
)

from .parsing import (  # noqa
    ParsedLine,
    parse_line,
    parse_line_,
    parse_line_strict,
)

from .rendering import (  # noqa
    RenderedLine,
    render_line,
    render_line_,
    render_line_strict,
)

from .tags import (  # noqa
    escape_tag_value,
    parse_tags,
    unescape_tag_value,
    validate_tag_name,
    validate_tag_value,
)



"""
TODO:
 - join
 - subquery
"""
import enum
import typing as ta

from ... import dataclasses as dc
from ... import lang
from .base import Node
from .exprs import CanExpr
from .exprs import Expr
from .idents import Ident
from .multi import MultiBuilder
from .names import CanName
from .names import Name


##


class Relation(Node, lang.Abstract):
    pass


#


class Table(Relation, lang.Final):
    n: Name
    a: Ident | None = dc.xfield(None, repr_fn=dc.opt_repr)


#


class JoinKind(enum.Enum):
    DEFAULT = enum.auto()
    INNER = enum.auto()
    LEFT = enum.auto()
    RIGHT = enum.auto()
    FULL = enum.auto()
    CROSS = enum.auto()
    NATURAL = enum.auto()


class Join(Relation, lang.Final):
    k: JoinKind
    l: Relation
    r: Relation

    c: Expr | None = None


##


CanTable: ta.TypeAlias = Table | CanName
CanRelation: ta.TypeAlias = Relation | CanTable


class RelationBuilder(MultiBuilder):
    def table(self, n: CanTable) -> Table:
        if isinstance(n, Table):
            return n
        else:
            return Table(self.name(n))

    #

    def join(
            self,
            k: JoinKind,
            l: CanRelation,
            r: CanRelation,
            *cs: CanExpr,
    ) -> Join:
        return Join(
            k,
            self.relation(l),
            self.relation(r),
            self.or_(*cs) if cs else None,
        )

    def default_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.DEFAULT, l, r, *cs)

    def inner_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.INNER, l, r, *cs)

    def left_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.LEFT, l, r, *cs)

    def right_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.RIGHT, l, r, *cs)

    def full_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.FULL, l, r, *cs)

    def cross_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.CROSS, l, r, *cs)

    def natural_join(self, l: CanRelation, r: CanRelation, *cs: CanExpr) -> Join:
        return self.join(JoinKind.NATURAL, l, r, *cs)

    #

    def relation(self, o: CanRelation) -> Relation:
        if isinstance(o, Relation):
            return o
        else:
            return self.table(o)



from ..codecs import make_object_lazy_loaded_codec
from ..codecs import make_str_object_codec
from .backends.default import dumps
from .backends.default import dumps_compact
from .backends.default import dumps_pretty
from .backends.default import loads


##


JSON_CODEC = make_str_object_codec('json', dumps, loads)

# @omlish-manifest
_JSON_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'JSON_CODEC', JSON_CODEC)

#

JSON_COMPACT_CODEC = make_str_object_codec('json-compact', dumps_compact, loads)

# @omlish-manifest
_JSON_COMPACT_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'JSON_COMPACT_CODEC', JSON_COMPACT_CODEC)

#

JSON_PRETTY_CODEC = make_str_object_codec('json-pretty', dumps_pretty, loads)

# @omlish-manifest
_JSON_PRETTY_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'JSON_PRETTY_CODEC', JSON_PRETTY_CODEC)



import typing as ta

from ..lite.maysyncs import Maywaitable
from ..lite.maysyncs import make_maysync as _make_maysync
from ..lite.maysyncs import maysync as _maysync
from .functions import as_async


T = ta.TypeVar('T')
P = ta.ParamSpec('P')

MaysyncP: ta.TypeAlias = ta.Callable[P, Maywaitable[T]]


##


def make_maysync(
        s: ta.Callable[P, T],
        a: ta.Callable[P, ta.Awaitable[T]] | None = None,
) -> MaysyncP[P, T]:
    return ta.cast('MaysyncP[P, T]', _make_maysync(
        s,
        a if a is not None else as_async(s),
    ))


def maysync(m: ta.Callable[P, ta.Awaitable[T]]) -> MaysyncP[P, T]:
    return ta.cast('MaysyncP[P, T]', _maysync(
        m,
    ))



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
# Based on bluelet ( https://github.com/sampsyo/bluelet ) by Adrian Sampson, original license:
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from .core import _CoreBlueletApi
from .files import _FilesBlueletApi
from .runner import _RunnerBlueletApi
from .sockets import _SocketsBlueletApi


class BlueletApi(
    _RunnerBlueletApi,
    _SocketsBlueletApi,
    _FilesBlueletApi,
    _CoreBlueletApi,
):
    pass


bluelet = BlueletApi()



# @omlish-lite
# ruff: noqa: UP006 UP007 UP043 UP045
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import email.message
import email.parser
import typing as ta

from ....lite.check import check
from .errors import CoroHttpClientErrors
from .io import CoroHttpClientIo


##


class CoroHttpClientHeaders:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    #

    MAX_HEADERS: ta.ClassVar[int] = 100

    @classmethod
    def read_headers(cls) -> ta.Generator[CoroHttpClientIo.Io, ta.Optional[bytes], ta.List[bytes]]:
        """
        Reads potential header lines into a list from a file pointer.

        Length of line is limited by MAX_LINE, and number of headers is limited by MAX_HEADERS.
        """

        headers = []
        while True:
            line = check.isinstance((yield CoroHttpClientIo.ReadLineIo(CoroHttpClientIo.MAX_LINE + 1)), bytes)
            if len(line) > CoroHttpClientIo.MAX_LINE:
                raise CoroHttpClientErrors.LineTooLongError(CoroHttpClientErrors.LineTooLongError.LineType.HEADER)

            headers.append(line)
            if len(headers) > cls.MAX_HEADERS:
                raise CoroHttpClientErrors.ClientError(f'got more than {cls.MAX_HEADERS} headers')

            if line in (b'\r\n', b'\n', b''):
                break

        return headers

    @classmethod
    def parse_header_lines(cls, header_lines: ta.Sequence[bytes]) -> email.message.Message:
        """
        Parses only RFC2822 headers from header lines.

        email Parser wants to see strings rather than bytes. But a TextIOWrapper around self.rfile would buffer too many
        bytes from the stream, bytes which we later need to read as bytes. So we read the correct bytes here, as bytes,
        for email Parser to parse.
        """

        text = b''.join(header_lines).decode('iso-8859-1')
        return email.parser.Parser().parsestr(text)

    @classmethod
    def parse_headers(cls) -> ta.Generator[CoroHttpClientIo.Io, ta.Optional[bytes], email.message.Message]:
        """Parses only RFC2822 headers from a file pointer."""

        headers = yield from cls.read_headers()
        return cls.parse_header_lines(headers)



# ruff: noqa: UP007 UP045
import dataclasses as dc
import typing as ta


##


@dc.dataclass(frozen=True)
class AttrRepr:
    attrs: ta.Sequence[str]

    # _: dc.KW_ONLY

    with_module: bool = False
    use_qualname: bool = False
    with_id: bool = False
    value_filter: ta.Optional[ta.Callable[[ta.Any], bool]] = None
    recursive: bool = False

    @classmethod
    def of(cls, *attrs: str, **kwargs: ta.Any) -> 'AttrRepr':
        return cls(attrs, **kwargs)

    #

    def _build_(self, obj: ta.Any) -> str:
        vs = ', '.join(
            f'{attr}={v!r}'
            for attr in self.attrs
            for v in [getattr(obj, attr)]
            if self.value_filter is None or self.value_filter(v)
        )

        return (
            f'{obj.__class__.__module__ + "." if self.with_module else ""}'
            f'{obj.__class__.__qualname__ if self.use_qualname else obj.__class__.__name__}'
            f'{("@" + hex(id(obj))[2:]) if self.with_id else ""}'
            f'({vs})'
        )

    _build: ta.ClassVar[ta.Callable[[ta.Any], str]]

    def __call__(self, obj: ta.Any) -> str:
        try:
            build: ta.Any = self._build

        except AttributeError:
            build = self._build_
            if self.recursive:
                build = self._reprlib().recursive_repr()(build)
            object.__setattr__(self, '_build', build)

        return build(obj)

    #

    def __get__(self, instance, owner):
        if instance is None:
            return self

        def __repr__(other):  # noqa
            return self(other)

        return __repr__.__get__(instance, owner)

    #

    _reprlib_: ta.ClassVar[ta.Any]

    @classmethod
    def _reprlib(cls) -> ta.Any:
        try:
            return cls._reprlib_
        except AttributeError:
            pass

        reprlib = __import__('reprlib')
        cls._reprlib_ = reprlib
        return reprlib


def attr_repr(obj: ta.Any, *attrs: str, **kwargs: ta.Any) -> str:
    return AttrRepr(attrs, **kwargs)(obj)



from .... import lang
from .base import KnownKeyword
from .base import StrKeyword
from .base import StrToKeywordsKeyword


##


class CoreKeyword(KnownKeyword, lang.Abstract, lang.Sealed):
    pass


##


class Id(StrKeyword, CoreKeyword, lang.Final, tag='$id'):
    pass


class SchemaKeyword(StrKeyword, CoreKeyword, lang.Final, tag='$schema'):
    pass


class Ref(StrKeyword, CoreKeyword, lang.Final, tag='$ref'):
    pass


class Defs(StrToKeywordsKeyword, CoreKeyword, lang.Final, tag='$defs', aliases=['definitions']):
    pass



import abc
import typing as ta

from .. import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')
MV = ta.TypeVar('MV', bound=ta.Iterable)


##


class MultiMap(ta.Mapping[K, MV], abc.ABC, ta.Generic[K, V, MV]):
    pass


SequenceMultiMap: ta.TypeAlias = MultiMap[K, V, ta.Sequence[V]]
AbstractSetMultiMap: ta.TypeAlias = MultiMap[K, V, ta.AbstractSet[V]]


##


class BiMultiMap(MultiMap[K, V, MV], abc.ABC, ta.Generic[K, V, MV]):
    @abc.abstractmethod
    def inverse(self) -> 'InverseBiMultiMap[K, V, MV]':
        raise NotImplementedError


class InverseBiMultiMap(ta.Mapping[V, K], abc.ABC, ta.Generic[K, V, MV]):
    @abc.abstractmethod
    def inverse(self) -> BiMultiMap[K, V, MV]:
        raise NotImplementedError


SequenceBiMultiMap: ta.TypeAlias = BiMultiMap[K, V, ta.Sequence[V]]
AbstractSetBiMultiMap: ta.TypeAlias = BiMultiMap[K, V, ta.AbstractSet[V]]


##


class InverseBiMultiMapImpl(InverseBiMultiMap[K, V, MV], ta.Generic[K, V, MV]):
    def __init__(self, m: BiMultiMap[K, V, MV], dct: ta.Mapping[V, K]) -> None:
        super().__init__()

        self._m = m
        self._dct = dct

    def inverse(self) -> BiMultiMap[K, V, MV]:
        return self._m

    def __getitem__(self, key: V, /) -> K:
        return self._dct[key]

    def __len__(self) -> int:
        return len(self._dct)

    def __iter__(self) -> ta.Iterator[V]:
        return iter(self._dct)


class BaseBiMultiMap(BiMultiMap[K, V, MV], abc.ABC, ta.Generic[K, V, MV]):
    def __init__(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__()

        dct: dict[K, MV] = {}
        i_dct: dict[V, K] = {}
        for k, mv in lang.yield_dict_init(*args, **kwargs):
            l: list[V] = []
            for v in mv:
                if v in i_dct:
                    raise KeyError(v)
                l.append(v)
                i_dct[v] = k
            dct[k] = self._aggregate_values(l)

        self._dct = dct
        self._i: InverseBiMultiMap[K, V, MV] = InverseBiMultiMapImpl(self, i_dct)

    @abc.abstractmethod
    def _aggregate_values(self, vs: list[V]) -> MV:
        raise NotImplementedError

    def inverse(self) -> 'InverseBiMultiMap[K, V, MV]':
        return self._i

    def __getitem__(self, key: K, /) -> MV:
        return self._dct[key]

    def __len__(self) -> int:
        return len(self._dct)

    def __iter__(self) -> ta.Iterator[K]:
        return iter(self._dct)


#


class TupleBiMultiMap(BaseBiMultiMap[K, V, tuple[V, ...]], ta.Generic[K, V]):
    def _aggregate_values(self, vs: list[V]) -> tuple[V, ...]:
        return tuple(vs)


# FIXME: lame
# lang.static_check_issubclass[BiMultiMap[int, str, tuple[str, ...]]](TupleBiMultiMap[int, str])
# lang.static_check_issubclass[BiMultiMap[int, str, ta.Sequence[str]]](TupleBiMultiMap[int, str])
# lang.static_check_issubclass[SequenceBiMultiMap[int, str]](TupleBiMultiMap[int, str])


@ta.overload
def seq_bi_multi_map(dct: ta.Mapping[K, ta.Iterable[V]]) -> SequenceBiMultiMap[K, V]:
    ...


@ta.overload
def seq_bi_multi_map(items: ta.Iterable[tuple[K, ta.Iterable[V]]]) -> SequenceBiMultiMap[K, V]:
    ...


def seq_bi_multi_map(*args, **kwargs):
    return TupleBiMultiMap(*args, **kwargs)


#


class FrozensetBiMultiMap(BaseBiMultiMap[K, V, frozenset[V]], ta.Generic[K, V]):
    def _aggregate_values(self, vs: list[V]) -> frozenset[V]:
        return frozenset(vs)


# FIXME: lame
# lang.static_check_issubclass[AbstractSetBiMultiMap[int, str]](FrozensetBiMultiMap[int, str])


@ta.overload
def abs_set_bi_multi_map(dct: ta.Mapping[K, ta.Iterable[V]]) -> AbstractSetBiMultiMap[K, V]:
    ...


@ta.overload
def abs_set_bi_multi_map(items: ta.Iterable[tuple[K, ta.Iterable[V]]]) -> AbstractSetBiMultiMap[K, V]:
    ...


def abs_set_bi_multi_map(*args, **kwargs):
    return FrozensetBiMultiMap(*args, **kwargs)



import typing as ta

from .identity import IdentityKeyDict
from .identity import IdentitySet


T = ta.TypeVar('T')


##


class RankedSeq(ta.Sequence[T]):
    def __init__(self, it: ta.Iterable[T], *, identity: bool = False) -> None:
        super().__init__()

        self._lst = list(it)
        self._ranks: ta.Mapping[T, int] = (
            (IdentityKeyDict if identity else dict)
            ((e, i) for i, e in enumerate(self._lst))
        )
        if len(self._ranks) != len(self._lst):
            raise ValueError(f'{len(self._ranks)} != {len(self._lst)}')

    @property
    def debug(self) -> ta.Sequence[T]:
        return self._lst

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._lst)

    def __getitem__(self, rank: int) -> T:  # type: ignore
        return self._lst[rank]

    def __len__(self) -> int:
        return len(self._lst)

    def __contains__(self, obj: T) -> bool:  # type: ignore
        return obj in self._ranks

    @property
    def ranks(self) -> ta.Mapping[T, int]:
        return self._ranks

    def rank(self, obj: T) -> int:
        return self._ranks[obj]


##


class RankedSetSeq(ta.Sequence[ta.AbstractSet[T]]):
    def __init__(self, it: ta.Iterable[ta.Iterable[T]], *, identity: bool = False) -> None:
        super().__init__()

        self._lst = [(IdentitySet if identity else set)(e) for e in it]
        self._ranks: ta.Mapping[T, int] = (
            (IdentityKeyDict if identity else dict)
            ((e, i) for i, es in enumerate(self._lst) for e in es)
        )
        if len(self._ranks) != sum(map(len, self._lst)):
            raise ValueError(f'{len(self._ranks)} != {sum(map(len, self._lst))}')

    @property
    def debug(self) -> ta.Sequence[ta.AbstractSet[T]]:
        return self._lst

    def __iter__(self) -> ta.Iterator[ta.AbstractSet[T]]:
        return iter(self._lst)

    def __getitem__(self, rank: int) -> ta.AbstractSet[T]:  # type: ignore
        return self._lst[rank]

    def __len__(self) -> int:
        return len(self._lst)

    def __contains__(self, obj: T) -> bool:  # type: ignore
        return obj in self._ranks

    @property
    def ranks(self) -> ta.Mapping[T, int]:
        return self._ranks

    def rank(self, obj: T) -> int:
        return self._ranks[obj]



# ruff: noqa: UP045
# @omlish-lite
"""
TODO:
 - __del__
"""
import abc
import functools
import typing as ta


T = ta.TypeVar('T')
T_co = ta.TypeVar('T_co', covariant=True)
_X = ta.TypeVar('_X')

_MaysyncGen = ta.Generator['_MaysyncOp', ta.Any, T]  # ta.TypeAlias


##


class Maywaitable(ta.Protocol[T_co]):
    def s(self) -> T_co:
        ...

    def a(self) -> ta.Awaitable[T_co]:
        ...

    def m(self) -> ta.Awaitable[T_co]:
        ...


Maysync = ta.Callable[..., Maywaitable[T]]  # ta.TypeAlias  # omlish-amalg-typing-no-move


class Maysync_(abc.ABC):  # noqa
    pass


##


class _Maywaitable(abc.ABC, ta.Generic[_X, T]):
    @ta.final
    def __init__(
            self,
            x: _X,
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        self.x = x
        self.args = args
        self.kwargs = kwargs

    @ta.final
    def m(self) -> ta.Awaitable[T]:
        return _MaysyncFuture(_MaysyncOp(
            ta.cast(ta.Any, self.x),
            *self.args,
            **self.kwargs,
        ))


##


@ta.final
class _FnMaysync(Maysync_, ta.Generic[T]):
    def __init__(
            self,
            s: ta.Callable[..., T],
            a: ta.Callable[..., ta.Awaitable[T]],
    ) -> None:
        if s is None:
            raise TypeError(s)
        if a is None:
            raise TypeError(a)
        self.s = s
        self.a = a

    def __get__(self, instance, owner=None):
        return _FnMaysync(
            self.s.__get__(instance, owner),  # noqa
            self.a.__get__(instance, owner),  # noqa
        )

    def __call__(self, *args, **kwargs):
        return _FnMaywaitable(self, *args, **kwargs)


@ta.final
class _FnMaywaitable(_Maywaitable[_FnMaysync[T], T]):
    def s(self) -> T:
        return self.x.s(*self.args, **self.kwargs)

    async def a(self) -> T:
        return await self.x.a(*self.args, **self.kwargs)


def make_maysync(
        s: ta.Callable[..., T],
        a: ta.Callable[..., ta.Awaitable[T]],
) -> Maysync[T]:
    return _FnMaysync(s, a)


##


@ta.final
class _MgMaysync(Maysync_, ta.Generic[T]):
    def __init__(
            self,
            mg: ta.Callable[..., _MaysyncGen[T]],
    ) -> None:
        self.mg = mg

    def __get__(self, instance, owner=None):
        return _MgMaysync(
            self.mg.__get__(instance, owner),  # noqa
        )

    def __call__(self, *args, **kwargs):
        return _MgMaywaitable(self, *args, **kwargs)


@ta.final
class _MgMaywaitable(_Maywaitable[_MgMaysync[T], T]):
    def s(self) -> T:
        g = self.x.mg(*self.args, **self.kwargs)

        i: ta.Any = None
        e: ta.Any = None

        while True:
            try:
                if e is not None:
                    o = g.throw(e)
                else:
                    o = g.send(i)
            except StopIteration as ex:
                return ex.value

            i = None
            e = None

            if not isinstance(o, _MaysyncOp):
                raise TypeError(o)
            try:
                i = o.x(*o.args, **o.kwargs).s()
            except BaseException as ex:  # noqa
                e = ex

            del o

    async def a(self) -> T:
        g = self.x.mg(*self.args, **self.kwargs)

        i: ta.Any = None
        e: ta.Any = None

        while True:
            try:
                if e is not None:
                    o = g.throw(e)
                else:
                    o = g.send(i)
            except StopIteration as ex:
                return ex.value

            i = None
            e = None

            if not isinstance(o, _MaysyncOp):
                raise TypeError(o)
            try:
                i = await o.x(*o.args, **o.kwargs).a()
            except BaseException as ex:  # noqa
                e = ex

            del o


@ta.final
class _MgMaysyncFn:
    def __init__(self, m):
        self.m = m

        functools.update_wrapper(self, m)

    def __get__(self, instance, owner=None):
        return _MgMaysyncFn(
            self.m.__get__(instance, owner),
        )

    def __call__(self, *args, **kwargs):
        a = self.m(*args, **kwargs).__await__()
        try:
            g = iter(a)
            try:
                while True:
                    try:
                        o = g.send(None)
                    except StopIteration as e:
                        return e.value

                    if not isinstance(o, _MaysyncFuture):
                        raise TypeError(o)
                    if not o.done:
                        try:
                            o.result = yield o.op
                        except BaseException as e:  # noqa
                            o.error = e
                        o.done = True

            finally:
                g.close()

        finally:
            a.close()


def maysync(m: ta.Callable[..., ta.Awaitable[T]]) -> Maysync[T]:
    return _MgMaysync(_MgMaysyncFn(m))


##


@ta.final
class _MaysyncOp:
    def __init__(
            self,
            x: Maysync[T],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        self.x = x
        self.args = args
        self.kwargs = kwargs


@ta.final
class _MaysyncFuture(ta.Generic[T]):
    def __init__(
            self,
            op: _MaysyncOp,
    ) -> None:
        self.op = op

    done: bool = False
    result: ta.Optional[T] = None
    error: ta.Optional[BaseException] = None

    def __await__(self):
        if not self.done:
            yield self
        if not self.done:
            raise RuntimeError("await wasn't used with event future")
        if self.error is not None:
            raise self.error
        return self.result



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - raw
 - blocks / inheritance
"""
import dataclasses as dc
import io
import re
import typing as ta

from ..funcs.builders import FnBuilder
from ..funcs.builders import SimpleFnBuilder
from ..lite.cached import cached_nullary
from ..lite.check import check
from ..lite.maybes import Maybe


MinjaTemplateFragmentKind = ta.Literal['expr', 'stmt', 'for', 'if', 'elif']  # ta.TypeAlias


##


@dc.dataclass(frozen=True)
class MinjaTemplateParam:
    name: str
    default: Maybe[ta.Any] = Maybe.empty()

    def __post_init__(self) -> None:
        check.arg(self.name.isidentifier())

    @classmethod
    def of(cls, obj: ta.Union[str, 'MinjaTemplateParam']) -> 'MinjaTemplateParam':
        if isinstance(obj, MinjaTemplateParam):
            return obj
        elif isinstance(obj, str):
            return MinjaTemplateParam.new(obj)
        else:
            raise TypeError(obj)

    @classmethod
    def new(cls, name: str, *defaults: ta.Any) -> 'MinjaTemplateParam':
        dfl: Maybe[ta.Any]
        if defaults:
            [dv] = defaults
            dfl = Maybe.just(dv)
        else:
            dfl = Maybe.empty()
        return cls(name, dfl)


class MinjaTemplateFn(ta.Protocol):
    def __call__(self, **kwargs: ta.Any) -> str:
        ...


class MinjaTemplate:
    def __init__(
            self,
            fn: MinjaTemplateFn,
            params: ta.Sequence[MinjaTemplateParam],
    ) -> None:
        super().__init__()

        self._fn = fn
        self._params = params

    @property
    def params(self) -> ta.Sequence[MinjaTemplateParam]:
        return self._params

    def __call__(self, **kwargs: ta.Any) -> str:
        return self._fn(**kwargs)


##


class MinjaTemplateCompiler:
    """
    Compiles a template string into a Python fn. The returned fn takes a dictionary 'context' and returns
    the rendered string.

    Supported syntax:
    - Literal text remains literal.
    - {{ expr }}: Evaluates 'expr' in the given context and writes its str() to output.
    - {% code %}: Executes the given Python code (e.g. 'for x in items:'), must be terminated appropriately.
    - {% endfor %} to close for loops.
    - {% endif %} to close if blocks.
    - {# comment #}: Ignored completely.
    """

    DEFAULT_INDENT: str = ' ' * 4

    def __init__(
            self,
            src: str,
            params: ta.Sequence[ta.Union[str, MinjaTemplateParam]],
            *,
            indent: str = DEFAULT_INDENT,
            fragment_processor: ta.Optional[ta.Callable[[MinjaTemplateFragmentKind, str], str]] = None,
            strict_strings: bool = False,
            stringifier: ta.Optional[ta.Callable[[ta.Any], str]] = None,
            fn_builder: ta.Optional[FnBuilder] = None,
    ) -> None:
        super().__init__()

        check.not_isinstance(params, str)

        self._src = check.isinstance(src, str)
        self._params = [
            MinjaTemplateParam.of(p)
            for p in params
        ]
        check.unique(p.name for p in self._params)
        self._indent_str: str = check.non_empty_str(indent)

        if fragment_processor is None:
            fragment_processor = lambda _, s: s  # noqa
        self._fragment_processor = fragment_processor

        if stringifier is None:
            if strict_strings:
                stringifier = self._strict_stringify
            else:
                stringifier = lambda o: str(o)
        self._stringifier = stringifier

        if fn_builder is None:
            fn_builder = SimpleFnBuilder()
        self._fn_builder = fn_builder

        self._stack: ta.List[ta.Literal['for', 'if']] = []

    @staticmethod
    def _strict_stringify(o: ta.Any) -> str:
        if not isinstance(o, str):
            raise TypeError(o)
        return o

    #

    _TAG_PAT = re.compile(
        r'({{.*?}}|{%.*?%}|{#.*?#})',
        flags=re.DOTALL,
    )

    @classmethod
    def _split_tags(cls, src: str) -> ta.List[ta.Tuple[str, str]]:
        raw = cls._TAG_PAT.split(src)

        #

        parts: ta.List[ta.Tuple[str, str]] = []
        for s in raw:
            if not s:
                continue

            for g, l, r in [
                ('{', '{{', '}}'),
                ('%', '{%', '%}'),
                ('#', '{#', '#}'),
            ]:
                if s.startswith(l) and s.endswith(r):
                    parts.append((g, s[len(l):-len(r)]))
                    break
            else:
                parts.append(('', s))

        #

        for i, (g, s) in enumerate(parts):
            if s.startswith('-'):
                if i > 0:
                    lg, ls = parts[i - 1]
                    parts[i - 1] = (lg, ls.rstrip())
                s = s[1:].lstrip()

            if s.endswith('-'):
                if i < len(parts) - 1:
                    rg, rs = parts[i + 1]
                    parts[i + 1] = (rg, rs.lstrip())
                s = s[:-1].rstrip()

            parts[i] = (g, s)

        #

        parts = [(g, s) for g, s in parts if g or s]

        #

        return parts

    #

    def _indent(self, line: str, ofs: int = 0) -> str:
        return self._indent_str * (len(self._stack) + 1 + ofs) + line

    #

    _RENDER_FN_NAME = '__minja__render'

    class Rendered(ta.NamedTuple):
        src: str
        ns: ta.Dict[str, ta.Any]

    @cached_nullary
    def render(self) -> Rendered:
        lines: ta.List[str] = []

        ns: ta.Dict[str, ta.Any] = {
            '__minja__StringIO': io.StringIO,
            '__minja__stringify': self._stringifier,
        }

        parts = self._split_tags(self._src)

        if not self._params:
            lines.append(f'def {self._RENDER_FN_NAME}():')
        else:
            lines.append(f'def {self._RENDER_FN_NAME}(')
            lines.append(self._indent('*,'))
            for p in self._params:
                if p.default.present:
                    check.not_in(p.name, ns)
                    dn = f'__minja__default__{p.name}'
                    ns[dn] = p.default.must()
                    lines.append(self._indent(f'{p.name}={dn},'))
                else:
                    lines.append(self._indent(f'{p.name},'))
            lines.append('):')

        lines.append(self._indent('__minja__output = __minja__StringIO()'))

        for g, s in parts:
            if g == '{':
                expr = s.strip()
                lines.append(self._indent(f'__minja__output.write(__minja__stringify({self._fragment_processor("expr", expr)}))'))  # noqa

            elif g == '%':
                stmt = s.strip()

                if stmt.startswith('for '):
                    lines.append(self._indent(f'for {self._fragment_processor("for", stmt[4:])}:'))
                    self._stack.append('for')
                elif stmt.startswith('endfor'):
                    check.equal(self._stack.pop(), 'for')

                elif stmt.startswith('if '):
                    lines.append(self._indent(f'if {self._fragment_processor("if", stmt[3:])}:'))
                    self._stack.append('if')
                elif stmt.startswith('elif '):
                    check.equal(self._stack[-1], 'if')
                    lines.append(self._indent(f'elif {self._fragment_processor("elif", stmt[5:])}:', -1))
                elif stmt.strip() == 'else':
                    check.equal(self._stack[-1], 'if')
                    lines.append(self._indent('else:', -1))
                elif stmt.strip() == 'endif':
                    check.equal(self._stack.pop(), 'if')

                else:
                    lines.append(self._indent(self._fragment_processor('stmt', stmt)))

            elif g == '#':
                pass

            elif not g:
                if s:
                    safe_text = s.replace('"""', '\\"""')
                    lines.append(self._indent(f'__minja__output.write("""{safe_text}""")'))

            else:
                raise KeyError(g)

        check.empty(self._stack)

        lines.append(self._indent('return __minja__output.getvalue()'))

        return self.Rendered('\n'.join(lines), ns)

    #

    def compile(
            self,
            ns: ta.Optional[ta.Mapping[str, ta.Any]] = None,
    ) -> MinjaTemplate:
        rendered = self.render()

        ns = dict(ns or {})
        for k, v in rendered.ns.items():
            if k in ns:
                raise KeyError(k)
            ns[k] = v

        render_fn = self._fn_builder.build_fn(
            self._RENDER_FN_NAME,
            rendered.src,
            ns,
        )

        return MinjaTemplate(
            ta.cast(MinjaTemplateFn, check.callable(render_fn)),
            self._params,
        )


##


def compile_minja_template(
        src: str,
        params: ta.Sequence[ta.Union[str, MinjaTemplateParam]] = (),
        *,
        ns: ta.Optional[ta.Mapping[str, ta.Any]] = None,
        **kwargs: ta.Any,
) -> MinjaTemplate:
    return MinjaTemplateCompiler(
        src,
        params,
        **kwargs,
    ).compile(ns)


def render_minja_template(src: str, **kwargs: ta.Any) -> str:
    tmpl = compile_minja_template(src, list(kwargs))
    return tmpl(**kwargs)



"""
TODO:
 - find_vars
 - some kind of EnvKey / EnvMarker(lang.Marker), namespace-able env keys - but how to weave these into tmpl srcs?
"""
import abc
import dataclasses as dc
import string
import typing as ta

from .. import lang
from .minja import MinjaTemplate
from .minja import MinjaTemplateParam
from .minja import compile_minja_template


if ta.TYPE_CHECKING:
    import jinja2
else:
    jinja2 = lang.proxy_import('jinja2')


##


class Templater(lang.Abstract):
    """
    This is named a 'templatER', not 'template', because it probably refers to a bigger thing itself called a
    'template' (like a `jinja2.Template`).
    """

    @dc.dataclass(frozen=True)
    class Context:
        env: ta.Mapping[str, ta.Any] | None = None

    @abc.abstractmethod
    def render(self, ctx: Context) -> str:
        raise NotImplementedError


def templater_context(
        *envs: ta.Mapping[str, ta.Any] | None,
        strict: bool = False,
) -> Templater.Context:
    env: dict[str, ta.Any] = {}
    for e in envs:
        if e is None:
            continue
        for k, v in e.items():
            if strict and k in env:
                raise KeyError(k)
            env[k] = v

    return Templater.Context(
        env=env or None,
    )


##


@dc.dataclass(frozen=True)
class FormatTemplater(Templater):
    """https://docs.python.org/3/library/string.html#format-specification-mini-language"""

    fmt: str

    def render(self, ctx: Templater.Context) -> str:
        return self.fmt.format(**(ctx.env or {}))


format_templater = FormatTemplater


##


@dc.dataclass(frozen=True)
class Pep292Templater(Templater):
    """https://peps.python.org/pep-0292/"""

    tmpl: string.Template

    def render(self, ctx: Templater.Context) -> str:
        return self.tmpl.substitute(ctx.env or {})

    @classmethod
    def from_string(cls, src: str) -> 'Pep292Templater':
        return cls(string.Template(src))


pep292_templater = Pep292Templater.from_string


##


@dc.dataclass(frozen=True)
class MinjaTemplater(Templater):
    tmpl: MinjaTemplate

    ENV_IDENT: ta.ClassVar[str] = 'env'

    def render(self, ctx: Templater.Context) -> str:
        return self.tmpl(**{self.ENV_IDENT: ctx.env or {}})

    @classmethod
    def from_string(
            cls,
            src: str,
            **ns: ta.Any,
    ) -> 'MinjaTemplater':
        tmpl = compile_minja_template(
            src,
            [
                MinjaTemplateParam(cls.ENV_IDENT),
                *[
                    MinjaTemplateParam.new(k, v)
                    for k, v in ns.items()
                ],
            ],
        )

        return cls(tmpl)


minja_templater = MinjaTemplater.from_string


##


@dc.dataclass(frozen=True)
class JinjaTemplater(Templater):
    tmpl: 'jinja2.Template'

    def render(self, ctx: Templater.Context) -> str:
        return self.tmpl.render(**(ctx.env or {}))

    @classmethod
    def from_string(
            cls,
            src: str,
            *,
            env: ta.Optional['jinja2.Environment'] = None,
            **kwargs: ta.Any,
    ) -> 'JinjaTemplater':
        if env is None:
            env = jinja2.Environment()
        tmpl = env.from_string(src)
        return cls(
            tmpl,
            **kwargs,
        )


jinja_templater = JinjaTemplater.from_string



"""
import sys, threading, pdb, functools
def _attach(repl):
    frame = sys._current_frames()[threading.enumerate()[0].ident]
    debugger = pdb.Pdb(
        stdin=repl.conn.makefile('r'),
        stdout=repl.conn.makefile('w'),
    )
    debugger.reset()
    while frame:
        frame.f_trace = debugger.trace_dispatch
        debugger.botframe = frame
        frame = frame.f_back
    debugger.set_step()
    frame.f_trace = debugger.trace_dispatch
"""
import ast
import codeop
import errno
import logging
import socket as sock
import sys
import threading
import traceback
import types
import typing as ta

from ... import check


log = logging.getLogger(__name__)


##


class DisconnectError(Exception):
    pass


class InteractiveSocketConsole:
    """code.InteractiveConsole but just different enough to not be worth subclassing."""

    ENCODING = 'utf-8'

    def __init__(
            self,
            conn: sock.socket,
            locals: dict[str, ta.Any] | None = None,  # noqa
            filename: str = '<console>',
    ) -> None:
        super().__init__()

        if locals is None:
            locals = {  # noqa
                '__name__': '__console__',
                '__doc__': None,
                '__console__': self,
            }

        self._conn = conn
        self._locals = locals
        self._filename = filename

        self._compiler = codeop.CommandCompiler()
        self._buffer: list[str] = []
        self._count = 0
        self._write_count = -1

    def reset_buffer(self) -> None:
        self._buffer = []

    @property
    def conn(self) -> sock.socket:
        return self._conn

    CPRT = 'Type "help", "copyright", "credits" or "license" for more information.'

    def interact(self, banner: str | None = None, exitmsg: str | None = None) -> None:
        log.info('Console %x on thread %r interacting', id(self), threading.current_thread().ident)

        try:
            ps1 = getattr(sys, 'ps1', '>>> ')
            ps2 = getattr(sys, 'ps2', '... ')

            if banner is None:
                self.write(f'Python {sys.version} on {sys.platform}\n{self.CPRT}\n({self.__class__.__name__})\n')
            elif banner:
                self.write(f'{banner!s}\n')

            more = False
            while True:
                try:
                    try:
                        line = self.raw_input(ps2 if more else ps1)
                    except EOFError:
                        self.write('\n')
                        break
                    else:
                        more = self.push_line(line)

                except KeyboardInterrupt:
                    self.write('\nKeyboardInterrupt\n')
                    self.reset_buffer()
                    more = False

            if exitmsg is None:
                self.write(f'now exiting {self.__class__.__name__}...\n')

            elif exitmsg != '':
                self.write(f'{exitmsg}\n')

        except DisconnectError:
            pass

        except OSError as oe:
            if oe.errno == errno.EBADF:
                pass

        finally:
            log.info('Console %x on thread %r finished', id(self), threading.current_thread().ident)

    def push_line(self, line: str) -> bool:
        self._buffer.append(line)
        source = '\n'.join(self._buffer)
        more = self.run_source(source, self._filename)
        if not more:
            self.reset_buffer()
        return more

    def raw_input(self, prompt: str = '') -> str:
        self.write(prompt)
        buf = b''
        while True:
            b = self._conn.recv(1)
            if not b:
                raise DisconnectError
            if b == b'\n':
                break
            buf += b
        return buf.decode(self.ENCODING)

    def write(self, data: str) -> None:
        self._conn.send(data.encode(self.ENCODING))

    def compile(
            self,
            source: str | ast.AST,
            filename: str = '<input>',
            symbol: str = 'single',
    ) -> types.CodeType | None:
        if isinstance(source, ast.AST):
            return self._compiler.compiler(source, filename, symbol)  # type: ignore
        else:
            return self._compiler(source, filename, symbol)

    def run_source(
            self,
            source: str | ast.AST,
            filename: str = '<input>',
            symbol: str = 'single',
    ) -> bool:
        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            # Case 1 (incorrect)
            self.show_syntax_error(filename)
            return False

        if code is None:
            # Case 2 (incomplete)
            return True

        # Case 3 (complete)
        try:
            node = ast.parse(source)  # type: ignore
        except (OverflowError, SyntaxError, ValueError):
            return True

        if (
                isinstance(node, ast.Module) and
                node.body and
                isinstance(node.body[-1], ast.Expr)
        ):
            expr = node.body[-1]
            source = ast.Interactive(
                [
                    *node.body[:-1],
                    ast.Assign(
                        [ast.Name(
                            f'_{self._count}',
                            ast.Store(),
                            lineno=expr.lineno,
                            col_offset=expr.col_offset,
                        )],
                        expr.value,
                        lineno=expr.lineno,
                        col_offset=expr.col_offset,
                    ),
                ],
            )
            ast.fix_missing_locations(source)
            self._write_count = self._count

        code = check.not_none(self.compile(source, filename, symbol))
        self.run_code(code)
        return False

    def run_code(self, code: types.CodeType) -> None:
        try:
            exec(code, self._locals)
        except SystemExit:
            raise
        except Exception:  # noqa
            self.show_traceback()
        else:
            if self._count == self._write_count:
                self.write(repr(self._locals[f'_{self._count}']))
                self.write('\n')
                self._count += 1

    def show_traceback(self) -> None:
        sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
        sys.last_traceback = last_tb
        try:
            lines = traceback.format_exception(ei[0], ei[1], last_tb.tb_next)  # type: ignore
            self.write(''.join(lines))
        finally:
            last_tb = ei = None  # type: ignore  # noqa

    def show_syntax_error(self, filename: str | None = None) -> None:
        et, e, tb = sys.exc_info()
        sys.last_type = et
        sys.last_value = e
        sys.last_traceback = tb
        if filename and et is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = e.args  # type: ignore
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                e = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = e
        lines = traceback.format_exception_only(et, e)
        self.write(''.join(lines))



import contextlib
import enum
import typing as ta

import pytest

from .... import check
from .... import inject as inj
from .... import lang
from .. import plugins


T = ta.TypeVar('T')


##


class PytestScope(enum.Enum):
    SESSION = enum.auto()
    PACKAGE = enum.auto()
    MODULE = enum.auto()
    CLASS = enum.auto()
    FUNCTION = enum.auto()


class Scopes(lang.Namespace, lang.Final):
    Session = inj.SeededScope(PytestScope.SESSION)
    Package = inj.SeededScope(PytestScope.PACKAGE)
    Module = inj.SeededScope(PytestScope.MODULE)
    Class = inj.SeededScope(PytestScope.CLASS)
    Function = inj.SeededScope(PytestScope.FUNCTION)


_SCOPES_BY_PYTEST_SCOPE: ta.Mapping[PytestScope, inj.SeededScope] = {
    check.isinstance(a.tag, PytestScope): a
    for n, a in Scopes.__dict__.items()
    if isinstance(a, inj.SeededScope)
}


##


_ACTIVE_HARNESSES: set['Harness'] = set()


class Harness:
    def __init__(self, es: inj.Elements) -> None:
        super().__init__()
        self._orig_es = es
        self._es = inj.as_elements(
            inj.bind(self),
            *[
                inj.as_elements(
                    inj.bind_scope(ss),
                    inj.bind_scope_seed(inj.Key(pytest.FixtureRequest, tag=pts), ss),
                )
                for pts, ss in _SCOPES_BY_PYTEST_SCOPE.items()
            ],
            es,
        )
        self._inj: inj.Injector | None = None

    ##

    @contextlib.contextmanager
    def activate(self) -> ta.Generator[ta.Self]:
        check.none(self._inj)
        check.not_in(self, _ACTIVE_HARNESSES)
        _ACTIVE_HARNESSES.add(self)
        try:
            with inj.create_managed_injector(self._es) as i:
                self._inj = i
                yield self
        finally:
            self._inj = None
            _ACTIVE_HARNESSES.remove(self)

    ##

    def __getitem__(
            self,
            target: inj.Key[T] | type[T],
    ) -> T:
        return check.not_none(self._inj)[target]

    def session(self) -> pytest.FixtureRequest:
        return self[inj.Key(pytest.FixtureRequest, tag=PytestScope.SESSION)]

    def package(self) -> pytest.FixtureRequest:
        return self[inj.Key(pytest.FixtureRequest, tag=PytestScope.PACKAGE)]

    def module(self) -> pytest.FixtureRequest:
        return self[inj.Key(pytest.FixtureRequest, tag=PytestScope.MODULE)]

    def class_(self) -> pytest.FixtureRequest:
        return self[inj.Key(pytest.FixtureRequest, tag=PytestScope.CLASS)]

    def function(self) -> pytest.FixtureRequest:
        return self[inj.Key(pytest.FixtureRequest, tag=PytestScope.FUNCTION)]

    ##

    @contextlib.contextmanager
    def _pytest_scope_manager(
            self,
            pytest_scope: PytestScope,
            request: pytest.FixtureRequest,
    ) -> ta.Generator[None]:
        ss = _SCOPES_BY_PYTEST_SCOPE[pytest_scope]
        with inj.enter_seeded_scope(check.not_none(self._inj), ss, {
            inj.Key(pytest.FixtureRequest, tag=pytest_scope): request,
        }):
            yield


##


@plugins.register
class HarnessPlugin:
    def __init__(self) -> None:
        super().__init__()
        self._harnesses_by_session: dict[ta.Any, Harness] = {}

    def get_session_harness(self, session: ta.Any) -> Harness:
        return self._harnesses_by_session[session]

    @pytest.fixture(scope='session', autouse=True)
    def _harness_scope_listener_session(self, request):
        with Harness(inj.as_elements(*_HARNESS_ELEMENTS_LIST)).activate() as harness:
            self._harnesses_by_session[request.session] = harness
            try:
                with harness._pytest_scope_manager(PytestScope.SESSION, request):  # noqa
                    yield
            finally:
                del self._harnesses_by_session[request.session]

    @pytest.fixture(scope='package', autouse=True)
    def _harness_scope_listener_package(self, request):
        harness = self.get_session_harness(request.session)
        with harness._pytest_scope_manager(PytestScope.PACKAGE, request):  # noqa
            yield

    @pytest.fixture(scope='module', autouse=True)
    def _harness_scope_listener_module(self, request):
        harness = self.get_session_harness(request.session)
        with harness._pytest_scope_manager(PytestScope.MODULE, request):  # noqa
            yield

    @pytest.fixture(scope='class', autouse=True)
    def _harness_scope_listener_class(self, request):
        harness = self.get_session_harness(request.session)
        with harness._pytest_scope_manager(PytestScope.CLASS, request):  # noqa
            yield

    @pytest.fixture(scope='function', autouse=True)  # noqa
    def _harness_scope_listener_function(self, request):
        harness = self.get_session_harness(request.session)
        with harness._pytest_scope_manager(PytestScope.FUNCTION, request):  # noqa
            yield


##


_HARNESS_ELEMENTS_LIST: list[inj.Elements] = []


def register(*args: inj.Element | inj.Elements) -> None:
    _HARNESS_ELEMENTS_LIST.append(inj.as_elements(*args))


def bind(
        scope: PytestScope | str = PytestScope.SESSION,
        *,
        eager: bool = False,  # FIXME
) -> ta.Callable[[T], T]:
    def inner(obj):
        pts = scope if isinstance(scope, PytestScope) else PytestScope[check.isinstance(scope, str).upper()]
        register(inj.as_elements(
            inj.bind(obj, in_=_SCOPES_BY_PYTEST_SCOPE[pts]),
        ))
        return obj
    return inner


@pytest.fixture
def harness(request) -> Harness:
    pm = request.session.config.pluginmanager
    hp = check.single(p for n, p in pm.list_name_plugin() if isinstance(p, HarnessPlugin))
    return hp.get_session_harness(request.session)



import collections.abc

from ... import check
from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactoryMatchClass
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactoryMatchClass
from .iterables import DEFAULT_ITERABLE_CONCRETE_TYPES
from .iterables import IterableMarshaler
from .iterables import IterableUnmarshaler


##


class SequenceNotStrMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Protocol) and rty.cls is lang.SequenceNotStr)
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Protocol)
        return IterableMarshaler(ctx.make(check.single(gty.args)))


class SequenceNotStrUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Protocol) and rty.cls is lang.SequenceNotStr)
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Protocol)
        cty = DEFAULT_ITERABLE_CONCRETE_TYPES[collections.abc.Sequence]  # type: ignore[type-abstract]
        return IterableUnmarshaler(cty, ctx.make(check.single(gty.args)))



# ruff: noqa: UP007 UP045
import contextlib
import dataclasses as dc
import signal
import sys
import threading
import typing as ta

from .. import check
from .. import lang
from .base import Bootstrap
from .base import ContextBootstrap
from .base import SimpleBootstrap


if ta.TYPE_CHECKING:
    import cProfile  # noqa
    import pstats

    from ..diag import debug as ddbg
    from ..diag import pycharm as diagpc
    from ..diag import replserver as diagrs
    from ..diag import threads as diagt

else:
    cProfile = lang.proxy_import('cProfile')  # noqa
    pstats = lang.proxy_import('pstats')

    ddbg = lang.proxy_import('..diag.debug', __package__)
    diagpc = lang.proxy_import('..diag.pycharm', __package__)
    diagrs = lang.proxy_import('..diag.replserver', __package__)
    diagt = lang.proxy_import('..diag.threads', __package__)


##


class CheckBootstrap(ContextBootstrap['CheckBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        breakpoint: bool = False

    @staticmethod
    def _breakpoint(exc: Exception) -> None:  # noqa
        breakpoint()  # noqa

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if not self._config.breakpoint:
            return

        check.register_on_raise(CheckBootstrap._breakpoint)
        try:
            yield
        finally:
            check.unregister_on_raise(CheckBootstrap._breakpoint)


##


class CprofileBootstrap(ContextBootstrap['CprofileBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        enable: bool = False
        builtins: bool = True

        outfile: ta.Optional[str] = None

        print: bool = False
        sort: str = 'cumtime'
        topn: int = 100

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if not self._config.enable:
            yield
            return

        prof = cProfile.Profile()
        prof.enable()
        try:
            yield

        finally:
            prof.disable()
            prof.create_stats()

            if self._config.print:
                pstats.Stats(prof) \
                    .strip_dirs() \
                    .sort_stats(self._config.sort) \
                    .print_stats(self._config.topn)

            if self._config.outfile is not None:
                prof.dump_stats(self._config.outfile)


##


class ThreadDumpBootstrap(ContextBootstrap['ThreadDumpBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        interval_s: ta.Optional[float] = None
        nodaemon: bool = False

        on_sigquit: bool = False

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if self._config.interval_s:
            tdt = diagt.create_thread_dump_thread(
                interval_s=self._config.interval_s,
                start=True,
                nodaemon=self._config.nodaemon,
            )
        else:
            tdt = None

        if self._config.on_sigquit:
            dump_threads_str = diagt.dump_threads_str

            def handler(signum, frame):
                print(dump_threads_str(), file=sys.stderr)

            prev_sq = lang.just(signal.signal(signal.SIGQUIT, handler))
        else:
            prev_sq = lang.empty()

        try:
            yield

        finally:
            if tdt is not None and not self._config.nodaemon:
                tdt.stop_nowait()

            if prev_sq.present:
                signal.signal(signal.SIGQUIT, prev_sq.must())


##


class TimebombBootstrap(ContextBootstrap['TimebombBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        delay_s: ta.Optional[float] = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if not self._config.delay_s:
            yield
            return

        tbt = diagt.create_timebomb_thread(
            self._config.delay_s,
            start=True,
        )
        try:
            yield
        finally:
            tbt.stop_nowait()


##


class PycharmBootstrap(SimpleBootstrap['PycharmBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        debug: ta.Optional[str] = None

    def run(self) -> None:
        if self._config.debug is not None:
            prd = diagpc.PycharmRemoteDebugger.parse(self._config.debug)
            diagpc.pycharm_remote_debugger_attach(prd)


##


class ReplServerBootstrap(ContextBootstrap['ReplServerBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        path: str | None = None

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if self._config.path is None:
            return

        with diagrs.ReplServer(diagrs.ReplServer.Config(
            path=self._config.path,
        )) as rs:
            thread = threading.Thread(target=rs.run, name='replserver')
            thread.start()

            yield


##


class DebugBootstrap(ContextBootstrap['DebugBootstrap.Config']):
    @dc.dataclass(frozen=True)
    class Config(Bootstrap.Config):
        enable: bool = False

    @contextlib.contextmanager
    def enter(self) -> ta.Iterator[None]:
        if not self._config.enable:
            return

        with ddbg.debugging_on_exception():
            yield



"""
TODO:
 - sync (lol) w/ asyncs.anyio
 - atomics
 - Once poison=False, PoisonedError
"""
import collections
import threading
import typing as ta

from . import lang


T = ta.TypeVar('T')


##


class Once:
    def __init__(self) -> None:
        super().__init__()
        self._done = False
        self._lock = threading.Lock()

    def do(self, fn: ta.Callable[[], None]) -> bool:
        if self._done:
            return False
        with self._lock:
            if self._done:
                return False  # type: ignore
            try:
                fn()
            finally:
                self._done = True
            return True


##


class Lazy(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()
        self._once = Once()
        self._v: lang.Maybe[T] = lang.empty()

    def peek(self) -> lang.Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = lang.just(v)

    def get(self, fn: ta.Callable[[], T]) -> T:
        def do():
            self._v = lang.just(fn())
        self._once.do(do)
        return self._v.must()


class LazyFn(ta.Generic[T]):
    def __init__(self, fn: ta.Callable[[], T]) -> None:
        super().__init__()
        self._fn = fn
        self._once = Once()
        self._v: lang.Maybe[T] = lang.empty()

    def peek(self) -> lang.Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = lang.just(v)

    def get(self) -> T:
        def do():
            self._v = lang.just(self._fn())
        self._once.do(do)
        return self._v.must()


##


class ConditionDeque(ta.Generic[T]):
    def __init__(
            self,
            *,
            cond: ta.Optional['threading.Condition'] = None,
            deque: collections.deque[T] | None = None,

            lock: ta.Optional['threading.RLock'] = None,
            maxlen: int | None = None,
            init: ta.Iterable[T] | None = None,
    ) -> None:
        super().__init__()

        if cond is None:
            cond = threading.Condition(lock=lock)
        if deque is None:
            deque = collections.deque(maxlen=maxlen)
        if init:
            deque.extend(init)

        self._cond = cond
        self._deque = deque

    @property
    def cond(self) -> 'threading.Condition':
        return self._cond

    @property
    def deque(self) -> collections.deque[T]:
        return self._deque

    def push(
            self,
            *items: T,
            n: int = 1,
    ) -> None:
        with self.cond:
            self.deque.extend(items)
            self.cond.notify(n)

    def pop(
            self,
            timeout: float | None = None,
            *,
            if_empty: ta.Callable[[], None] | None = None,
    ) -> T:
        with self.cond:
            if not self.deque and if_empty is not None:
                if_empty()
            while not self.deque:
                self.cond.wait(timeout)
            return self.deque.popleft()



"""
TODO:
 - Maysync impls?
 - base Protocol so adapters and real sa impls can be used interchangeably (if in asyncio ctx)?
"""
import contextlib
import typing as ta

import sqlalchemy as sa
import sqlalchemy.ext.asyncio as saa

from ...asyncs import all as au


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


AsyncEngineLike: ta.TypeAlias = ta.Union[saa.AsyncEngine, 'AsyncEngine']
AsyncConnectionLike: ta.TypeAlias = ta.Union[saa.AsyncConnection, 'AsyncConnection']
AsyncTransactionLike: ta.TypeAlias = ta.Union[saa.AsyncTransaction, 'AsyncTransaction']


##


class AsyncTransaction:
    def __init__(self, underlying: saa.AsyncTransaction) -> None:
        super().__init__()
        self._underlying = underlying

    @property
    def underlying(self) -> saa.AsyncTransaction:
        return self._underlying

    ##

    @au.mark_asyncio
    async def close(self) -> None:
        await au.from_asyncio(self._underlying.close)()

    @au.mark_asyncio
    async def rollback(self) -> None:
        await au.from_asyncio(self._underlying.rollback)()

    @au.mark_asyncio
    async def commit(self) -> None:
        await au.from_asyncio(self._underlying.commit)()


class AsyncConnection:
    def __init__(self, underlying: saa.AsyncConnection) -> None:
        super().__init__()
        self._underlying = underlying

    @property
    def underlying(self) -> saa.AsyncConnection:
        return self._underlying

    ##

    @contextlib.asynccontextmanager
    @au.mark_asyncio
    async def begin(self) -> ta.AsyncIterator[AsyncTransaction]:
        async with au.from_asyncio_context(self._underlying.begin()) as u:
            yield AsyncTransaction(u)

    @au.mark_asyncio
    async def execute(
            self,
            statement: ta.Any,
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> sa.CursorResult[ta.Any]:
        return await au.from_asyncio(self._underlying.execute)(statement, *args, **kwargs)

    @au.mark_asyncio
    async def run_sync(
            self,
            fn: ta.Callable[ta.Concatenate[sa.Connection, P], T],
            *args: P.args,
            **kwargs: P.kwargs,
    ) -> T:
        return await au.from_asyncio(self._underlying.run_sync)(fn, *args, **kwargs)


class AsyncEngine:
    def __init__(self, underlying: saa.AsyncEngine) -> None:
        super().__init__()
        self._underlying = underlying

    @property
    def underlying(self) -> saa.AsyncEngine:
        return self._underlying

    ##

    @contextlib.asynccontextmanager
    @au.mark_asyncio
    async def connect(self) -> ta.AsyncIterator[AsyncConnection]:
        async with au.from_asyncio_context(self._underlying.connect()) as u:
            yield AsyncConnection(u)

    @au.mark_asyncio
    async def dispose(self, close: bool = True) -> None:
        await au.from_asyncio(self._underlying.dispose)(close)


##


@ta.overload
def async_adapt(obj: AsyncEngine) -> AsyncEngine:
    ...


@ta.overload
def async_adapt(obj: AsyncConnection) -> AsyncConnection:
    ...


@ta.overload
def async_adapt(obj: AsyncTransaction) -> AsyncTransaction:
    ...


@ta.overload
def async_adapt(obj: saa.AsyncEngine) -> AsyncEngine:
    ...


@ta.overload
def async_adapt(obj: saa.AsyncConnection) -> AsyncConnection:
    ...


@ta.overload
def async_adapt(obj: saa.AsyncTransaction) -> AsyncTransaction:
    ...


def async_adapt(obj: ta.Any) -> ta.Any:
    if isinstance(obj, (AsyncEngine, AsyncConnection, AsyncTransaction)):
        return obj
    if isinstance(obj, saa.AsyncTransaction):
        return AsyncTransaction(obj)
    if isinstance(obj, saa.AsyncConnection):
        return AsyncConnection(obj)
    if isinstance(obj, saa.AsyncEngine):
        return AsyncEngine(obj)
    raise TypeError(obj)



"""
This is all gross, but this business always tends to be, and it at least centralizes and abstracts away all the
grossness and hides it behind a small and stable interface. It supports exactly as much as is needed by the codebase at
any given moment. It is 90% of why the codebase is strictly 3.13+ - I've tried to maintain backwards compat with this
kind of thing many times before and it's not worth my time, especially with lite code as an alternative.

I'm exploring extracting and distilling down mypy's type system to replace all of this, both to add some formalism /
give it some guiding North Star to make all of its decisions for it, and to add some long sought capabilities (infer,
meet, join, solve, ...), but it's quite a bit of work and not a priority at the moment.

TODO:
 - visitor / transformer
 - uniform collection isinstance - items() for mappings, iter() for other
 - also check instance type in isinstance not just items lol
 - ta.Generic in mro causing trouble - omit? no longer 1:1
 - cache this shit, esp generic_mro shit
  - cache __hash__ in Generic/Union
"""
import abc
import dataclasses as dc
import types
import typing as ta


_NoneType = types.NoneType  # type: ignore

_NONE_TYPE_FROZENSET: frozenset['Type'] = frozenset([_NoneType])


_AnnotatedAlias = ta._AnnotatedAlias  # type: ignore  # noqa
_CallableGenericAlias = ta._CallableGenericAlias  # type: ignore  # noqa
_GenericAlias = ta._GenericAlias  # type: ignore  # noqa
_LiteralGenericAlias = ta._LiteralGenericAlias  # type: ignore  # noqa
_ProtocolMeta = ta._ProtocolMeta  # noqa
_SpecialGenericAlias = ta._SpecialGenericAlias  # type: ignore  # noqa
_UnionGenericAlias = ta._UnionGenericAlias   # type: ignore  # noqa


##


_Protocol = getattr(ta, 'Protocol')
_Generic = getattr(ta, 'Generic')

if not isinstance(_Protocol, type) or not issubclass(_Protocol, _Generic):
    raise TypeError(f'typing.Protocol is not a proper typing.Generic subtype')


##


@dc.dataclass(frozen=True)
class _Special:
    name: str
    alias: _SpecialGenericAlias  # type: ignore
    origin: type
    nparams: int


_KNOWN_SPECIALS = [
    _Special(
        v._name,  # noqa
        v,
        v.__origin__,
        v._nparams,  # noqa
    )
    for v in ta.__dict__.values()  # noqa
    if isinstance(v, _SpecialGenericAlias)
]

_KNOWN_SPECIALS_BY_NAME = {s.name: s for s in _KNOWN_SPECIALS}
_KNOWN_SPECIALS_BY_ALIAS = {s.alias: s for s in _KNOWN_SPECIALS}
_KNOWN_SPECIALS_BY_ORIGIN = {s.origin: s for s in _KNOWN_SPECIALS}

_MAX_KNOWN_SPECIAL_TYPE_VARS = 16

_KNOWN_SPECIAL_TYPE_VARS = tuple(
    ta.TypeVar(f'_{i}')  # noqa
    for i in range(_MAX_KNOWN_SPECIAL_TYPE_VARS)
)


##


_SIMPLE_GENERIC_ALIAS_TYPES: set[type] = set()


def is_simple_generic_alias_type(oty: type) -> bool:
    return (
        oty is _GenericAlias or
        oty is ta.GenericAlias or  # type: ignore  # noqa
        oty in _SIMPLE_GENERIC_ALIAS_TYPES
    )


def get_params(obj: ta.Any) -> tuple[ta.TypeVar, ...]:
    if isinstance(obj, type):
        if issubclass(obj, ta.Generic):  # type: ignore
            return obj.__dict__.get('__parameters__', ())  # noqa

        if (ks := _KNOWN_SPECIALS_BY_ORIGIN.get(obj)) is not None:
            if (np := ks.nparams) < 0:
                raise TypeError(obj)
            return _KNOWN_SPECIAL_TYPE_VARS[:np]

    oty = type(obj)

    if is_simple_generic_alias_type(oty):
        return obj.__dict__.get('__parameters__', ())  # noqa

    if oty is _CallableGenericAlias:
        raise NotImplementedError('get_params not yet implemented for typing.Callable')

    raise TypeError(obj)


def _is_immediate_protocol(obj: ta.Any) -> bool:
    return isinstance(obj, _ProtocolMeta) and obj.__dict__['_is_protocol']


def is_union_type(cls: ta.Any) -> bool:
    if hasattr(ta, 'UnionType'):
        return ta.get_origin(cls) in {ta.Union, getattr(ta, 'UnionType')}

    else:
        return ta.get_origin(cls) in {ta.Union}


def get_orig_bases(obj: ta.Any) -> ta.Any:
    return obj.__orig_bases__  # noqa


def get_orig_class(obj: ta.Any) -> ta.Any:
    return obj.__orig_class__  # noqa


def get_newtype_supertype(obj: ta.Any) -> ta.Any:
    return obj.__supertype__


def get_type_var_bound(obj: ta.Any) -> ta.Any:
    return obj.__bound__


##


class TypeInfo(abc.ABC):  # noqa
    pass


Type: ta.TypeAlias = ta.Union[  # noqa
    type,
    ta.TypeVar,
    TypeInfo,
]


TYPES: tuple[type, ...] = (
    type,
    ta.TypeVar,
    TypeInfo,
)


##


@dc.dataclass(frozen=True)
class Union(TypeInfo):
    args: frozenset[Type]

    @property
    def is_optional(self) -> bool:
        return _NoneType in self.args

    def without_none(self) -> Type:
        if _NoneType not in self.args:
            return self
        rem = self.args - _NONE_TYPE_FROZENSET
        if len(rem) == 1:
            return next(iter(rem))
        return Union(rem)


#


GenericLikeCls = ta.TypeVar('GenericLikeCls')


@dc.dataclass(frozen=True)
class GenericLike(TypeInfo, abc.ABC, ta.Generic[GenericLikeCls]):
    cls: GenericLikeCls

    # args and params are the same length - params maps to the generic origin's params:
    # args   : map[int, V] = (int, V) | map[T, T] = (T, T)
    # params : map[int, V] = (_0, _1) | map[T, T] = (_0, _1)
    args: tuple[Type, ...]
    params: tuple[ta.TypeVar, ...] = dc.field(compare=False, repr=False)

    obj: ta.Any = dc.field(compare=False, repr=False)

    def __post_init__(self) -> None:
        if not isinstance(self.cls, type):
            raise ReflectTypeError(f'GenericLike {self.cls=} must be a type')
        if len(self.args) != len(self.params):
            raise ReflectTypeError(f'GenericLike {self.args=} must be same length as {self.params=}')

    def full_eq(self, other: 'GenericLike') -> bool:
        return (
            type(self) is type(other) and
            self.cls == other.cls and
            self.args == other.args and
            self.params == other.params and
            self.obj == other.obj
        )


@dc.dataclass(frozen=True)
class Generic(GenericLike[type]):
    pass


@dc.dataclass(frozen=True)
class Protocol(GenericLike[ta.Any]):
    # cls will still be a type - it will be the topmost _is_protocol=True class.
    # it may however be ta.Protocol, which *is* a type, but not according to mypy.
    pass


#


@dc.dataclass(frozen=True)
class NewType(TypeInfo):
    obj: ta.Any
    ty: Type


#


@dc.dataclass(frozen=True)
class Annotated(TypeInfo):
    ty: Type
    md: ta.Sequence[ta.Any]

    obj: ta.Any = dc.field(compare=False, repr=False)


#


@dc.dataclass(frozen=True)
class Literal(TypeInfo):
    args: tuple[ta.Any, ...]

    obj: ta.Any = dc.field(compare=False, repr=False)


#


class Any(TypeInfo):
    pass


ANY = Any()


##


class ReflectTypeError(TypeError):
    pass


class Reflector:
    def __init__(
            self,
            *,
            override: ta.Callable[[ta.Any], ta.Any] | None = None,
    ) -> None:
        super().__init__()

        self._override = override

    #

    def is_type(self, obj: ta.Any) -> bool:
        try:
            self._type(obj, check_only=True)
        except ReflectTypeError:
            return False
        else:
            return True

    def type(self, obj: ta.Any) -> Type:
        if (ty := self._type(obj, check_only=False)) is None:
            raise RuntimeError(obj)
        return ty

    #

    def _type(self, obj: ta.Any, *, check_only: bool) -> Type | None:
        ##
        # Overrides beat everything

        if self._override is not None:
            if (ovr := self._override(obj)) is not None:
                return ovr

        ##
        # Any

        if obj is ta.Any:
            return ANY

        ##
        # Already a Type?

        if isinstance(obj, (ta.TypeVar, TypeInfo)):  # noqa
            return obj

        oty = type(obj)

        ##
        # Union

        if oty is _UnionGenericAlias or oty is types.UnionType:
            if check_only:
                return None

            return Union(frozenset(self.type(a) for a in ta.get_args(obj)))

        ##
        # NewType

        if isinstance(obj, ta.NewType):  # noqa
            if check_only:
                return None

            return NewType(obj, get_newtype_supertype(obj))

        ##
        # Simple Generic

        if (
                is_simple_generic_alias_type(oty) or
                oty is _CallableGenericAlias
        ):
            if check_only:
                return None

            origin = ta.get_origin(obj)

            args = ta.get_args(obj)

            if origin is ta.Protocol:
                params = get_params(obj)

            elif oty is _CallableGenericAlias:
                p, r = args
                if p is Ellipsis or isinstance(p, ta.ParamSpec):
                    raise ReflectTypeError(f'Callable argument not yet supported for {obj=}')
                args = (*p, r)
                params = _KNOWN_SPECIAL_TYPE_VARS[:len(args)]

            elif origin is tuple:
                params = _KNOWN_SPECIAL_TYPE_VARS[:len(args)]

            elif origin is ta.Generic:
                params = args

            else:
                params = get_params(origin)

            if len(args) != len(params):
                raise ReflectTypeError(f'Mismatched {args=} and {params=} for {obj=}')

            if not isinstance(origin, type):
                raise ReflectTypeError(f'Generic origin {origin!r} is not a type')

            if origin is ta.Protocol:
                if args != params:
                    raise ReflectTypeError(f'Protocol argument not yet supported for {args=}, {params=}')
                return Protocol(
                    ta.Protocol,
                    args,
                    params,
                    obj,
                )

            r_args = tuple(self.type(a) for a in args)

            if _is_immediate_protocol(origin):
                return Protocol(
                    origin,
                    r_args,
                    params,
                    obj,
                )

            return Generic(
                origin,
                r_args,
                params,
                obj,
            )

        ##
        # Full Generic

        if isinstance(obj, type):
            if check_only:
                return None

            if _is_immediate_protocol(obj):
                params = get_params(obj)
                return Protocol(
                    obj,
                    params,
                    params,
                    obj,
                )

            if issubclass(obj, ta.Generic):  # type: ignore
                params = get_params(obj)
                if params:
                    return Generic(
                        obj,
                        params,
                        params,
                        obj,
                    )

            return obj

        ##
        # Special Generic

        if isinstance(obj, _SpecialGenericAlias):
            if (ks := _KNOWN_SPECIALS_BY_ALIAS.get(obj)) is not None:
                if check_only:
                    return None

                if (np := ks.nparams) < 0:
                    raise ReflectTypeError(obj)

                params = _KNOWN_SPECIAL_TYPE_VARS[:np]
                return Generic(
                    ks.origin,
                    params,
                    params,
                    obj,
                )

        ##
        # Annotated

        if isinstance(obj, _AnnotatedAlias):
            if check_only:
                return None

            o = ta.get_args(obj)[0]
            return Annotated(self.type(o), md=obj.__metadata__, obj=obj)

        ##
        # Literal

        if isinstance(obj, _LiteralGenericAlias):
            if check_only:
                return None

            return Literal(ta.get_args(obj), obj=obj)

        ##
        # Failure

        raise ReflectTypeError(obj)


DEFAULT_REFLECTOR = Reflector()

is_type = DEFAULT_REFLECTOR.is_type
type_ = DEFAULT_REFLECTOR.type



import dataclasses as dc
import typing as ta


##


def fields_dict(cls_or_instance: ta.Any) -> dict[str, dc.Field]:
    return {f.name: f for f in dc.fields(cls_or_instance)}


##


def iter_items(obj: ta.Any) -> ta.Iterator[tuple[str, ta.Any]]:
    for f in dc.fields(obj):
        yield (f.name, getattr(obj, f.name))


def iter_keys(obj: ta.Any) -> ta.Iterator[str]:
    for f in dc.fields(obj):
        yield f.name


def iter_values(obj: ta.Any) -> ta.Iterator[ta.Any]:
    for f in dc.fields(obj):
        yield getattr(obj, f.name)



from .base import (  # noqa
    Builder,
    HasQn,
    Node,
    NodeComparisonTypeError,
    Value,
)

from .binary import (  # noqa
    Binary,
    BinaryBuilder,
    BinaryOp,
    BinaryOps,
)

from .exprs import (  # noqa
    CanExpr,
    CanLiteral,
    Expr,
    ExprBuilder,
    Literal,
    NameExpr,
    ParamExpr,
)

from .idents import (  # noqa
    CanIdent,
    Ident,
    IdentBuilder,
    IdentLike,
)

from .multi import (  # noqa
    Multi,
    MultiBuilder,
    MultiKind,
)

from .names import (  # noqa
    CanName,
    Name,
    NameBuilder,
    NameLike,
)

from .ops import (  # noqa
    OpKind,
)

from .params import (  # noqa
    CanParam,
    Param,
    ParamBuilder,
    as_param,
)

from .relations import (  # noqa
    CanRelation,
    CanTable,
    Join,
    JoinKind,
    Relation,
    RelationBuilder,
    Table,
)

from .rendering import (  # noqa
    Renderer,
    StdRenderer,
    render,
)

from .selects import (  # noqa
    AllSelectItem,
    CanRelation,
    ExprSelectItem,
    Select,
    SelectBuilder,
    SelectItem,
)

from .std import (  # noqa
    StdBuilder,
)

from .stmts import (  # noqa
    CanExpr,
    ExprStmt,
    Stmt,
    StmtBuilder,
)

from .unary import (  # noqa
    Unary,
    UnaryBuilder,
    UnaryOp,
    UnaryOps,
)

from .unions import (  # noqa
    Union,
    UnionBuilder,
)


##


Q = StdBuilder()


##


from ...lang.imports import _register_conditional_import  # noqa

_register_conditional_import('...marshal', '.marshal', __package__)


##


from ...lang.imports import _trigger_conditional_imports  # noqa

_trigger_conditional_imports(__package__)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import re
import typing as ta

from ..lite.check import check
from ..subprocesses.run import SubprocessRun
from ..subprocesses.run import SubprocessRunnable
from ..subprocesses.run import SubprocessRunOutput


##


DEFAULT_DOCKER_NS1_RUN_IMAGE: str = 'debian'


def build_docker_ns1_run_args(
        *cmd: str,
        image: ta.Optional[str] = None,
        nsenter: str = 'nsenter',
) -> ta.List[str]:
    """
    - https://gist.github.com/BretFisher/5e1a0c7bcca4c735e716abf62afad389
    - https://github.com/justincormack/nsenter1/blob/8d3ba504b2c14d73c70cf34f1ec6943c093f1b02/nsenter1.c

    alt:
     - nc -U ~/Library/Containers/com.docker.docker/Data/debug-shell.sock
    """

    return [
        '--privileged',
        '--pid=host',
        (image if image is not None else DEFAULT_DOCKER_NS1_RUN_IMAGE),

        nsenter,
        '-t', '1',

        '-m',  # mount
        '-u',  # uts
        '-i',  # ipc
        '-n',  # net
        '-p',  # pid
        '-C',  # cgroup
        # '-U',  # user
        '-T',  # time

        *cmd,
    ]


def build_docker_ns1_run_cmd(
        *cmd: str,
        exe: str = 'docker',
        run_args: ta.Optional[ta.Sequence[str]] = None,
        **kwargs: ta.Any,
) -> ta.List[str]:
    if run_args is not None:
        check.not_isinstance(run_args, str)

    return [
        exe,
        'run',
        '--rm',
        *(run_args if run_args is not None else []),
        '-i',
        *build_docker_ns1_run_args(*cmd, **kwargs),
    ]


##


@dc.dataclass(frozen=True)
class DockerNs1ListUsedTcpPortsCommand(SubprocessRunnable[ta.List[int]]):
    kwargs: ta.Optional[ta.Mapping[str, str]] = None

    def make_run(self) -> SubprocessRun:
        return SubprocessRun.of(
            *build_docker_ns1_run_cmd(
                'netstat', '-tan',
            ),
            stdout='pipe',
            stderr='devnull',
            check=True,
        )

    _NETSTAT_LINE_PAT: ta.ClassVar[re.Pattern] = re.compile(r'\d{1,3}(.(\d{1,3})){3}:(?P<port>\d+)')

    def handle_run_output(self, output: SubprocessRunOutput) -> ta.List[int]:
        lines = [s for l in check.not_none(output.stdout).decode().splitlines() if (s := l.strip())]
        return [
            int(m.groupdict()['port'])
            for l in lines
            if len(ps := l.split(maxsplit=4)) > 3
            if (m := self._NETSTAT_LINE_PAT.fullmatch(ps[3])) is not None
        ]



import typing as ta

from .. import cached
from .. import dataclasses as dc
from .. import lang
from .base import Lifecycle


AbstractLifecycleT = ta.TypeVar('AbstractLifecycleT', bound='AbstractLifecycle')


##


class AbstractLifecycle(lang.Abstract):
    @dc.dataclass(frozen=True)
    class _Lifecycle(Lifecycle, lang.Final, ta.Generic[AbstractLifecycleT]):
        obj: AbstractLifecycleT

        def lifecycle_construct(self) -> None:
            self.obj._lifecycle_construct()  # noqa

        def lifecycle_start(self) -> None:
            self.obj._lifecycle_start()  # noqa

        def lifecycle_stop(self) -> None:
            self.obj._lifecycle_stop()  # noqa

        def lifecycle_destroy(self) -> None:
            self.obj._lifecycle_destroy()  # noqa

    @cached.property
    def _lifecycle(self) -> _Lifecycle[ta.Self]:
        return AbstractLifecycle._Lifecycle(self)

    def _lifecycle_construct(self) -> None:
        pass

    def _lifecycle_start(self) -> None:
        pass

    def _lifecycle_stop(self) -> None:
        pass

    def _lifecycle_destroy(self) -> None:
        pass



# ruff: noqa: UP006 UP045
# @omlish-lite
import abc
import io
import os.path
import sys
import typing as ta

from ..lite.cached import cached_nullary
from ..lite.check import check
from ..os.paths import is_path_in_dir


##


class FnBuilder(abc.ABC):
    @abc.abstractmethod
    def build_fn(
            self,
            name: str,
            src: str,
            ns: ta.Optional[ta.Mapping[str, ta.Any]] = None,
    ) -> ta.Callable:
        ...


#


class SimpleFnBuilder(FnBuilder):
    def build_fn(
            self,
            name: str,
            src: str,
            ns: ta.Optional[ta.Mapping[str, ta.Any]] = None,
    ) -> ta.Callable:
        ns = dict(ns or {})
        exec(src, ns)
        return ns[name]


build_fn = SimpleFnBuilder().build_fn


#


class DebugFnBuilder(FnBuilder):
    def __init__(
            self,
            *,
            mod_name_prefix: ta.Optional[str] = None,
            src_dir: ta.Optional[str] = None,
    ) -> None:
        super().__init__()

        if mod_name_prefix is None:
            mod_name_prefix = f'_{self.__class__.__name__}_{id(self):x}_'
        self._mod_name_prefix = mod_name_prefix

        self._given_src_dir = src_dir

        self._num_fns = 0
        self._mod_names: ta.Set[str] = set()
        self._installed_sys_path = False

    @cached_nullary
    def _src_dir(self) -> str:
        if self._given_src_dir is not None:
            return self._given_src_dir
        else:
            return __import__('tempfile').mkdtemp(prefix=f'_{self.__class__.__name__}_{os.getpid()}__')  # noqa

    @cached_nullary
    def _install_sys_path(self) -> None:
        if (src_dir := self._src_dir()) not in sys.path:
            self._installed_sys_path = True
            sys.path.append(src_dir)

    def uninstall_sys_path(self) -> None:
        if self._installed_sys_path:
            while True:
                try:
                    sys.path.remove(self._src_dir())
                except ValueError:
                    break
            self._installed_sys_path = True

    def _gen_mod_name(
            self,
            fn_num: int,
            fn_name: str,
            *,
            suffix: ta.Optional[str] = None,
    ) -> str:
        mod_name = f'{self._mod_name_prefix}{fn_num}__{fn_name}{suffix or ""}'

        check.not_in(mod_name, self._mod_names)
        check.not_in(mod_name, sys.modules)

        self._mod_names.add(mod_name)

        return mod_name

    def _build_mod(
            self,
            mod_name: str,
            mod_src: str,
    ) -> ta.Any:
        self._install_sys_path()

        src_dir = self._src_dir()
        src_file = os.path.join(src_dir, f'{mod_name}.py')
        check.state(is_path_in_dir(src_dir, src_file))

        with open(src_file, 'w') as f:
            f.write(mod_src)

        mod = __import__(mod_name)

        check.equal(mod.__file__, src_file)

        return mod

    def build_fn(
            self,
            name: str,
            src: str,
            ns: ta.Optional[ta.Mapping[str, ta.Any]] = None,
    ) -> ta.Callable:
        fn_num = self._num_fns
        self._num_fns += 1

        mod_name = self._gen_mod_name(fn_num, name)

        src_preamble: ta.List[str] = []

        if ns:
            ns_mod_name = self._gen_mod_name(fn_num, name, suffix='__ns')
            ns_mod = self._build_mod(ns_mod_name, '')

            for k, v in ns.items():
                setattr(ns_mod, k, v)

            src_preamble.append(f'from {ns_mod_name} import (')
            for k in sorted(ns):
                src_preamble.append(f'    {k},')
            src_preamble.append(')')

        src_sb = io.StringIO()
        if src_preamble:
            src_sb.write('\n'.join(src_preamble))
            src_sb.write('\n\n')
        src_sb.write(src)
        if not src.endswith('\n'):
            src_sb.write('\n')

        mod = self._build_mod(mod_name, src_sb.getvalue())

        fn = mod.__dict__[name]
        return fn



# ruff: noqa: UP045
# @omlish-lite
import contextlib
import os
import typing as ta


##


class EnvVar:
    def __init__(self, name: str) -> None:
        super().__init__()

        if not isinstance(name, str):
            raise TypeError(name)
        if not name:
            raise NameError(name)
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def __init_subclass__(cls, **kwargs):
        raise TypeError

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._name!r})'

    def __hash__(self) -> int:
        return hash((self.__class__, self._name))

    def __eq__(self, other):
        if not isinstance(other, EnvVar):
            return NotImplemented
        return self._name == other._name

    def __bool__(self) -> bool:
        raise TypeError

    def __str__(self) -> str:
        raise TypeError

    #

    @classmethod
    def _get_mut_environ(
            cls,
            environ: ta.Optional[ta.MutableMapping[str, str]] = None,
    ) -> ta.MutableMapping[str, str]:
        if environ is None:
            return os.environ
        return environ

    @classmethod
    def _get_environ(
            cls,
            environ: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> ta.Mapping[str, str]:
        if environ is None:
            return cls._get_mut_environ()
        return environ

    #

    def is_set(self, environ: ta.Optional[ta.Mapping[str, str]] = None) -> bool:
        return self._name in self._get_environ(environ)

    _NO_DEFAULT: ta.ClassVar[ta.Any] = object()

    @ta.overload
    def get(
            self,
            *,
            environ: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> str:
        ...

    @ta.overload
    def get(
            self,
            default: str,
            /,
            *,
            environ: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> str:
        ...

    @ta.overload
    def get(
            self,
            default: ta.Optional[str],
            /,
            *,
            environ: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> ta.Optional[str]:
        ...

    @ta.overload
    def get(
            self,
            default: ta.Callable[[], str],
            /,
            *,
            environ: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> str:
        ...

    @ta.overload
    def get(
            self,
            default: ta.Callable[[], ta.Optional[str]],
            /,
            *,
            environ: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> ta.Optional[str]:
        ...

    def get(self, default=_NO_DEFAULT, /, *, environ=None):
        environ = self._get_environ(environ)
        try:
            return environ[self._name]
        except KeyError:
            if default is self._NO_DEFAULT:
                raise
            elif callable(default):
                return default()
            else:
                return default

    #

    def set(self, value: ta.Optional[str], *, environ: ta.Optional[ta.MutableMapping[str, str]] = None) -> None:
        environ = self._get_mut_environ(environ)
        if value is not None:
            environ[self._name] = value
        else:
            del environ[self._name]

    @contextlib.contextmanager
    def set_context(
            self,
            value: ta.Optional[str],
            *,
            environ: ta.Optional[ta.MutableMapping[str, str]] = None,
    ) -> ta.Iterator[ta.Optional[str]]:
        prev = self.get(None, environ=environ)
        self.set(value, environ=environ)
        try:
            yield prev
        finally:
            self.set(prev, environ=environ)



import typing as ta

from ... import check
from ... import lang
from ... import marshal as msh
from .keywords.base import Keywords
from .keywords.parse import KeywordParser
from .keywords.render import render_keywords


##


class _KeywordsMarshaler(msh.Marshaler):
    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        return render_keywords(check.isinstance(o, Keywords))


class _KeywordsUnmarshaler(msh.Unmarshaler):
    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        return KeywordParser(allow_unknown=True).parse_keywords(check.isinstance(v, ta.Mapping))


@lang.static_init
def _install_standard_marshalling() -> None:
    msh.install_standard_factories(
        msh.TypeMapMarshalerFactory({Keywords: _KeywordsMarshaler()}),
        msh.TypeMapUnmarshalerFactory({Keywords: _KeywordsUnmarshaler()}),
    )



# While the library as a whole only supports modern python, the 'lite' package is importable and usable down to python
# 3.8. As a result this file needs to be kept basically empty as anything it would otherwise do would likely involve
# non-lite code.
